from typing import Any, Dict, List, Optional
from urllib.parse import quote

from .utils import request


class Repository:
    """Class for interacting with GitHub repositories"""

    def __init__(self, owner: str, repo: str):
        """
        Initialize repository client

        Args:
            owner: Repository owner
            repo: Repository name
        """
        if not owner or not repo:
            raise ValueError("Owner and repository name are required")
        if not isinstance(owner, str) or not isinstance(repo, str):
            raise ValueError("Owner and repository name must be strings")

        self.owner = owner.strip()
        self.repo = repo.strip()
        self.base_url = f"https://api.github.com/repos/{self.owner}/{self.repo}"

    def get_info(self) -> Dict[str, Any]:
        """Get basic repository information"""
        return request(self.base_url)

    def get_commits(self, page: int = 1, per_page: int = 30, sha: str = None) -> List[Dict[str, Any]]:
        """
        Get repository commits

        Args:
            page: Page number
            per_page: Items per page
            sha: SHA or branch name to start listing commits from
        """
        url = f"{self.base_url}/commits?page={page}&per_page={per_page}"
        if sha:
            url += f"&sha={quote(sha, safe='')}"
        return request(url)

    def get_commit(self, sha: str) -> Dict[str, Any]:
        """Get a specific commit details"""
        url = f"{self.base_url}/commits/{quote(sha, safe='')}"
        return request(url)

    def get_pull_requests(self, state: str = "all", page: int = 1, per_page: int = 30) -> List[Dict[str, Any]]:
        """
        Get repository pull requests

        Args:
            state: State of PRs to return (open/closed/all)
            page: Page number
            per_page: Items per page
        """
        url = f"{self.base_url}/pulls?state={state}&page={page}&per_page={per_page}"
        return request(url)

    def get_pull_request(self, pr_number: int) -> Dict[str, Any]:
        """Get specific pull request details"""
        url = f"{self.base_url}/pulls/{pr_number}"
        return request(url)

    def get_pull_request_commits(self, pr_number: int) -> List[Dict[str, Any]]:
        """Get commits in a specific pull request"""
        url = f"{self.base_url}/pulls/{pr_number}/commits"
        return request(url)

    def get_pull_request_files(self, pr_number: int) -> List[Dict[str, Any]]:
        """Get files changed in a specific pull request"""
        url = f"{self.base_url}/pulls/{pr_number}/files"
        return request(url)

    def get_issues(self, state: str = "all", page: int = 1, per_page: int = 30, labels: str = None) -> List[Dict[str, Any]]:
        """
        Get repository issues

        Args:
            state: State of issues to return (open/closed/all)
            page: Page number
            per_page: Items per page
            labels: Comma-separated list of label names
        """
        url = f"{self.base_url}/issues?state={state}&page={page}&per_page={per_page}"
        if labels:
            url += f"&labels={quote(labels, safe='')}"
        return request(url)

    def get_issue(self, issue_number: int) -> Dict[str, Any]:
        """Get specific issue details"""
        url = f"{self.base_url}/issues/{issue_number}"
        return request(url)

    def get_issue_comments(self, issue_number: int) -> List[Dict[str, Any]]:
        """Get comments on a specific issue"""
        url = f"{self.base_url}/issues/{issue_number}/comments"
        return request(url)

    def get_labels(self) -> List[Dict[str, Any]]:
        """Get repository labels"""
        url = f"{self.base_url}/labels"
        return request(url)

    def get_milestones(self, state: str = "all") -> List[Dict[str, Any]]:
        """Get repository milestones"""
        url = f"{self.base_url}/milestones?state={state}"
        return request(url)

    def get_contributors(self, page: int = 1, per_page: int = 30) -> List[Dict[str, Any]]:
        """Get repository contributors"""
        url = f"{self.base_url}/contributors?page={page}&per_page={per_page}"
        return request(url)

    def get_languages(self) -> Dict[str, int]:
        """Get repository language breakdown"""
        url = f"{self.base_url}/languages"
        return request(url)

    def get_teams(self) -> List[Dict[str, Any]]:
        """Get teams with access to repository"""
        url = f"{self.base_url}/teams"
        return request(url)

    def get_tags(self) -> List[Dict[str, Any]]:
        """Get repository tags"""
        url = f"{self.base_url}/tags"
        return request(url)

    def get_releases(self, page: int = 1, per_page: int = 30) -> List[Dict[str, Any]]:
        """Get repository releases"""
        url = f"{self.base_url}/releases?page={page}&per_page={per_page}"
        return request(url)

    def get_release(self, release_id: str) -> Dict[str, Any]:
        """Get specific release details"""
        url = f"{self.base_url}/releases/{release_id}"
        return request(url)

    def get_latest_release(self) -> Dict[str, Any]:
        """Get latest release"""
        url = f"{self.base_url}/releases/latest"
        return request(url)

    def get_branches(self, page: int = 1, per_page: int = 30) -> List[Dict[str, Any]]:
        """Get repository branches"""
        url = f"{self.base_url}/branches?page={page}&per_page={per_page}"
        return request(url)

    def get_branch(self, branch: str) -> Dict[str, Any]:
        """Get specific branch details"""
        url = f"{self.base_url}/branches/{quote(branch, safe='')}"
        return request(url)

    def get_contents(self, path: str = "", ref: Optional[str] = None) -> Dict[str, Any]:
        """Get contents of file or directory"""
        url = f"{self.base_url}/contents/{quote(path)}"
        if ref:
            url += f"?ref={quote(ref, safe='')}"
        return request(url)

    def get_collaborators(self) -> List[Dict[str, Any]]:
        """Get repository collaborators"""
        url = f"{self.base_url}/collaborators"
        return request(url)

    def get_workflows(self) -> List[Dict[str, Any]]:
        """Get repository GitHub Actions workflows"""
        url = f"{self.base_url}/actions/workflows"
        return request(url)

    def get_workflow_runs(self, workflow_id: str) -> List[Dict[str, Any]]:
        """Get workflow runs for a specific workflow"""
        url = f"{self.base_url}/actions/workflows/{workflow_id}/runs"
        return request(url)

    def get_deployments(self) -> List[Dict[str, Any]]:
        """Get repository deployments"""
        url = f"{self.base_url}/deployments"
        return request(url)

    def get_traffic(self) -> Dict[str, Any]:
        """Get repository traffic data"""
        url = f"{self.base_url}/traffic/views"
        return request(url)

    def get_code_frequency(self) -> List[List[int]]:
        """Get weekly commit activity"""
        url = f"{self.base_url}/stats/code_frequency"
        return request(url)

    def get_commit_activity(self) -> List[Dict[str, Any]]:
        """Get last year of commit activity"""
        url = f"{self.base_url}/stats/commit_activity"
        return request(url)

    def get_community_profile(self) -> Dict[str, Any]:
        """Get community profile metrics"""
        url = f"{self.base_url}/community/profile"
        return request(url)

    def get_readme(self, ref: Optional[str] = None) -> Dict[str, Any]:
        """
        Get repository README.

        Args:
            ref: The name of the commit/branch/tag (default: repo default branch)

        Returns:
            README content with encoding and download_url
        """
        url = f"{self.base_url}/readme"
        if ref:
            url += f"?ref={quote(ref, safe='')}"
        return request(url)

    def get_license(self) -> Dict[str, Any]:
        """
        Get repository license.

        Returns:
            License information including name, key, spdx_id, and content
        """
        url = f"{self.base_url}/license"
        return request(url)

    def get_topics(self) -> Dict[str, Any]:
        """
        Get repository topics.

        Returns:
            Dict with 'names' key containing list of topic strings
        """
        url = f"{self.base_url}/topics"
        return request(url)

    def get_forks(self, sort: str = "newest", page: int = 1, per_page: int = 30) -> List[Dict[str, Any]]:
        """
        List repository forks.

        Args:
            sort: Sort by (newest, oldest, stargazers, watchers)
            page: Page number
            per_page: Results per page (max 100)

        Returns:
            List of forked repositories
        """
        url = f"{self.base_url}/forks?sort={sort}&page={page}&per_page={per_page}"
        return request(url)

    def get_stargazers(self, page: int = 1, per_page: int = 30) -> List[Dict[str, Any]]:
        """
        List users who have starred the repository.

        Args:
            page: Page number
            per_page: Results per page (max 100)

        Returns:
            List of users who starred the repo
        """
        url = f"{self.base_url}/stargazers?page={page}&per_page={per_page}"
        return request(url)

    def get_watchers(self, page: int = 1, per_page: int = 30) -> List[Dict[str, Any]]:
        """
        List users watching the repository (subscribers).

        Args:
            page: Page number
            per_page: Results per page (max 100)

        Returns:
            List of users watching the repo
        """
        url = f"{self.base_url}/subscribers?page={page}&per_page={per_page}"
        return request(url)

    def compare(self, base: str, head: str) -> Dict[str, Any]:
        """
        Compare two commits, branches, or tags.

        Args:
            base: Base commit/branch/tag
            head: Head commit/branch/tag

        Returns:
            Comparison data including commits, files, and diff stats
        """
        url = f"{self.base_url}/compare/{quote(base, safe='')}...{quote(head, safe='')}"
        return request(url)

    def get_events(self, page: int = 1, per_page: int = 30) -> List[Dict[str, Any]]:
        """
        List repository events.

        Args:
            page: Page number
            per_page: Results per page (max 100)

        Returns:
            List of repository events
        """
        url = f"{self.base_url}/events?page={page}&per_page={per_page}"
        return request(url)



import json
from typing import Any, Dict, Generator, Optional, Union

from curl_cffi import CurlError
from curl_cffi.requests import Session

from webscout import exceptions
from webscout.AIbase import Provider, Response
from webscout.AIutel import AwesomePrompts, Conversation, Optimizers


class Llama3Mitril(Provider):
    """
    A class to interact with the Llama3 Mitril API. Implements the WebScout provider interface.
    """
    required_auth = False
    def __init__(
        self,
        is_conversation: bool = True,
        max_tokens: int = 2048,
        timeout: int = 30,
        intro: Optional[str] = None,
        filepath: Optional[str] = None,
        update_file: bool = True,
        proxies: dict = {},
        history_offset: int = 10250,
        act: Optional[str] = None,
        system_prompt: str = "You are a helpful, respectful and honest assistant.",
        temperature: float = 0.8,
    ):
        """Initializes the Llama3Mitril API."""
        self.session = Session()
        self.is_conversation = is_conversation
        self.max_tokens = max_tokens
        self.temperature = temperature
        self.api_endpoint = "https://llama3.mithrilsecurity.io/generate_stream"
        self.timeout = timeout
        self.last_response = {}
        self.system_prompt = system_prompt
        self.headers = {
            "Content-Type": "application/json",
            "DNT": "1",
        }
        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )
        Conversation.intro = (
            AwesomePrompts().get_act(
                act, raise_not_found=True, default=None, case_insensitive=True
            )
            if act
            else intro or Conversation.intro
        )
        self.conversation = Conversation(
            is_conversation, self.max_tokens, filepath, update_file
        )
        self.conversation.history_offset = history_offset
        # Update curl_cffi session headers and proxies
        self.session.headers.update(self.headers)
        self.session.proxies = proxies

    def _format_prompt(self, prompt: str) -> str:
        """Format the prompt for the Llama3 model"""
        return (
            f"<|begin_of_text|>"
            f"<|start_header_id|>system<|end_header_id|>{self.system_prompt}<|eot_id|>"
            f"<|start_header_id|>user<|end_header_id|>{prompt}<|eot_id|>"
            f"<|start_header_id|>assistant<|end_header_id|><|eot_id|>"
            f"<|start_header_id|>assistant<|end_header_id|>"
        )

    def ask(
        self,
        prompt: str,
        stream: bool = True,  # API supports streaming
        raw: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Response:
        """Sends a prompt to the Llama3 Mitril API and returns the response."""
        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            if optimizer in self.__available_optimizers:
                conversation_prompt = getattr(Optimizers, optimizer)(
                    conversation_prompt if conversationally else prompt
                )
            else:
                raise exceptions.FailedToGenerateResponseError(
                    f"Optimizer is not one of {self.__available_optimizers}"
                )

        data = {
            "inputs": self._format_prompt(conversation_prompt),
            "parameters": {
                "max_new_tokens": self.max_tokens,
                "temperature": self.temperature,
                "return_full_text": False
            }
        }

        def for_stream():
            streaming_response = ""  # Initialize outside try block
            try:
                # Use curl_cffi session post with impersonate
                response = self.session.post(
                    self.api_endpoint,
                    # headers are set on the session
                    json=data,
                    stream=True,
                    timeout=self.timeout,
                    # proxies are set on the session
                    impersonate="chrome110"  # Use a common impersonation profile
                )
                response.raise_for_status()  # Check for HTTP errors

                # Iterate over bytes and decode manually
                for line_bytes in response.iter_lines():
                    if line_bytes:
                        try:
                            line = line_bytes.decode('utf-8')
                            if line.startswith('data: '):
                                chunk_str = line.split('data: ', 1)[1]
                                chunk = json.loads(chunk_str)
                                if token_text := chunk.get('token', {}).get('text'):
                                    if '<|eot_id|>' not in token_text:
                                        streaming_response += token_text
                                        resp = {"text": token_text}
                                        # Yield dict or raw string chunk
                                        yield resp if not raw else token_text
                        except (json.JSONDecodeError, IndexError, UnicodeDecodeError):
                            # Ignore errors in parsing specific lines
                            continue

                # Update history after stream finishes
                self.last_response = {"text": streaming_response}
                self.conversation.update_chat_history(
                    prompt, streaming_response
                )

            except CurlError as e:  # Catch CurlError
                raise exceptions.FailedToGenerateResponseError(f"Request failed (CurlError): {e}") from e
            except Exception as e:  # Catch other potential exceptions (like HTTPError)
                err_text = getattr(e, 'response', None) and getattr(e.response, 'text', '')
                raise exceptions.FailedToGenerateResponseError(f"Failed to generate response ({type(e).__name__}): {e} - {err_text}") from e

        def for_non_stream():
            # Aggregate the stream using the updated for_stream logic
            full_response_text = ""
            try:
                # Ensure raw=False so for_stream yields dicts
                for chunk_data in for_stream():
                    if isinstance(chunk_data, dict) and "text" in chunk_data:
                        full_response_text += chunk_data["text"]
                    # Handle raw string case if raw=True was passed
                    elif raw and isinstance(chunk_data, str):
                         full_response_text += chunk_data
            except Exception as e:
                 # If aggregation fails but some text was received, use it. Otherwise, re-raise.
                 if not full_response_text:
                     raise exceptions.FailedToGenerateResponseError(f"Failed to get non-stream response: {str(e)}") from e

            # last_response and history are updated within for_stream
            # Return the final aggregated response dict or raw string
            return full_response_text if raw else self.last_response

        return for_stream() if stream else for_non_stream()

    def chat(
        self,
        prompt: str,
        stream: bool = True,  # Default to True as API supports it
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Union[str, Generator[str, None, None]]:
        """Generates a response from the Llama3 Mitril API."""

        def for_stream_chat():
            # ask() yields dicts or strings when streaming
            gen = self.ask(
                prompt, stream=True, raw=False,  # Ensure ask yields dicts
                optimizer=optimizer, conversationally=conversationally
            )
            for response_dict in gen:
                yield self.get_message(response_dict)  # get_message expects dict

        def for_non_stream_chat():
            # ask() returns dict or str when not streaming
            response_data = self.ask(
                prompt, stream=False, raw=False,  # Ensure ask returns dict
                optimizer=optimizer, conversationally=conversationally
            )
            return self.get_message(response_data)  # get_message expects dict

        return for_stream_chat() if stream else for_non_stream_chat()

    def get_message(self, response: Dict[str, Any]) -> str:
        """Extracts the message from the API response."""
        assert isinstance(response, dict), "Response should be of dict data-type only"
        return response["text"]


if __name__ == "__main__":
    # Ensure curl_cffi is installed
    from rich import print

    ai = Llama3Mitril(
        max_tokens=2048,
        temperature=0.8,
        timeout=30
    )

    for response in ai.chat("Hello", stream=True):
        print(response, end="", flush=True)



"""YouTube Hashtag functionality."""
import re
from typing import Any, Dict, List
from urllib.parse import quote

from .patterns import _ExtraPatterns as Patterns
from .utils import dup_filter, request


class Hashtag:
    """Class for YouTube hashtag operations."""

    BASE_URL = "https://www.youtube.com/hashtag"

    @staticmethod
    def get_videos(tag: str, limit: int = 20) -> List[str]:
        """
        Get videos associated with a hashtag.

        Args:
            tag: Hashtag (with or without #)
            limit: Maximum number of videos to return

        Returns:
            List of video IDs
        """
        if not tag:
            return []

        # Remove # if present and clean the tag
        tag = tag.lstrip('#').strip().lower()
        tag = re.sub(r'[^a-zA-Z0-9]', '', tag)

        if not tag:
            return []

        url = f"{Hashtag.BASE_URL}/{quote(tag)}"

        try:
            html = request(url)
            video_ids = Patterns.video_id.findall(html)
            return dup_filter(video_ids, limit)
        except Exception:
            return []

    @staticmethod
    def get_metadata(tag: str) -> Dict[str, Any]:
        """
        Get metadata about a hashtag.

        Args:
            tag: Hashtag (with or without #)

        Returns:
            Dictionary with hashtag info (name, video_count if available)
        """
        if not tag:
            return {}

        tag = tag.lstrip('#').strip().lower()
        tag = re.sub(r'[^a-zA-Z0-9]', '', tag)

        if not tag:
            return {}

        url = f"{Hashtag.BASE_URL}/{quote(tag)}"

        try:
            html = request(url)

            # Try to extract video count if available
            video_count_match = re.search(r'"videoCountText":\s*\{\s*"runs":\s*\[\s*\{\s*"text":\s*"([^"]+)"', html)
            video_count = video_count_match.group(1) if video_count_match else None

            # Get sample of videos
            video_ids = Patterns.video_id.findall(html)

            return {
                'tag': tag,
                'url': url,
                'video_count': video_count,
                'sample_videos': dup_filter(video_ids, 10)
            }
        except Exception:
            return {'tag': tag, 'url': url}

    @staticmethod
    def extract_from_text(text: str) -> List[str]:
        """
        Extract hashtags from text.

        Args:
            text: Text containing hashtags

        Returns:
            List of hashtags found
        """
        if not text:
            return []

        pattern = r'#([a-zA-Z0-9_]+)'
        matches = re.findall(pattern, text)
        return list(dict.fromkeys(matches))  # Remove duplicates, preserve order


if __name__ == "__main__":
    print("Testing Hashtag.get_videos:")
    videos = Hashtag.get_videos("python", 5)
    for vid in videos:
        print(f"  - {vid}")

    print("\nHashtag metadata:")
    meta = Hashtag.get_metadata("coding")
    print(f"  Tag: {meta.get('tag')}")
    print(f"  Videos: {len(meta.get('sample_videos', []))}")

    print("\nExtract hashtags:")
    text = "Check out my new #python #tutorial for #beginners!"
    tags = Hashtag.extract_from_text(text)
    print(f"  Found: {tags}")



"""
TempMail Package - Temporary Email Generation Functionality
Part of Webscout Extra tools
"""

from .base import (
    AsyncTempMailProvider,
    TempMailProvider,
    get_disposable_email,
    get_provider,
    get_random_email,
)
from .emailnator import EmailnatorProvider
from .mail_tm import MailTM, MailTMAsync
from .temp_mail_io import TempMailIO, TempMailIOAsync

__all__ = [
    'TempMailProvider',
    'AsyncTempMailProvider',
    'MailTM',
    'MailTMAsync',
    'TempMailIO',
    'TempMailIOAsync',
    'EmailnatorProvider',
    'get_random_email',
    'get_disposable_email',
    'get_provider'
]



from typing import Any, Dict, Generator, Optional, Union

from curl_cffi import CurlError  # Import CurlError

# Replace requests with curl_cffi
from curl_cffi.requests import Session  # Import Session

from webscout import exceptions
from webscout.AIbase import Provider, Response
from webscout.AIutel import AwesomePrompts, Conversation, Optimizers  # Import sanitize_stream
from webscout.litagent import LitAgent


class Netwrck(Provider):
    """
    A class to interact with the Netwrck.com API. Supports streaming.
    """
    greeting = """Hello! I'm a helpful assistant. How can I help you today?"""
    required_auth = False
    AVAILABLE_MODELS = [
        "thedrummer/valkyrie-49b-v1",
        "thedrummer/skyfall-36b-v2",
        "sao10k/l3-euryale-70b",
        "deepseek/deepseek-chat",
        "deepseek/deepseek-r1",
        "gryphe/mythomax-l2-13b",
        "nvidia/llama-3.1-nemotron-70b-instruct",
    ]

    def __init__(
        self,
        model: str = "deepseek/deepseek-r1",
        is_conversation: bool = True,
        max_tokens: int = 4096, # Note: max_tokens is not used by this API
        timeout: int = 30,
        intro: Optional[str] = None,
        filepath: Optional[str] = None,
        update_file: bool = False,
        proxies: Optional[dict] = None,
        history_offset: int = 0,
        act: Optional[str] = None,
        system_prompt: str = "You are a helpful assistant.",
        temperature: float = 0.7, # Note: temperature is not used by this API
        top_p: float = 0.8 # Note: top_p is not used by this API
    ):
        """Initializes the Netwrck API client."""
        if model not in self.AVAILABLE_MODELS:
            raise ValueError(f"Invalid model: {model}. Choose from: {self.AVAILABLE_MODELS}")

        self.model = model
        self.model_name = model
        self.system_prompt = system_prompt
        # Initialize curl_cffi Session
        self.session = Session()
        self.is_conversation = is_conversation
        self.max_tokens_to_sample = max_tokens
        self.timeout = timeout
        self.last_response: Dict[str, Any] = {}
        self.temperature = temperature
        self.top_p = top_p

        self.agent = LitAgent() # Keep for potential future use or other headers
        self.headers = {
            'authority': 'netwrck.com',
            'accept': '*/*',
            'accept-language': 'en-US,en;q=0.9',
            'content-type': 'application/json',
            'origin': 'https://netwrck.com',
            'referer': 'https://netwrck.com/',
            'user-agent': self.agent.random()
            # Add sec-ch-ua headers if needed for impersonation consistency
        }

        # Update curl_cffi session headers and proxies
        self.session.headers.update(self.headers)
        self.proxies = proxies or {}
        self.session.proxies = self.proxies # Assign proxies directly

        Conversation.intro = (
            AwesomePrompts().get_act(act, raise_not_found=True, default=None, case_insensitive=True)
            if act
            else intro or Conversation.intro
        )

        self.conversation = Conversation(is_conversation, max_tokens, filepath, update_file)
        self.conversation.history_offset = history_offset
        self.__available_optimizers = (
            method for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )

    @staticmethod
    def _netwrck_extractor(chunk: Union[str, Dict[str, Any]]) -> Optional[str]:
        """Removes surrounding quotes and handles potential escapes."""
        if isinstance(chunk, str):
            text = chunk.strip('"')
            # Handle potential unicode escapes if they appear
            # text = text.encode().decode('unicode_escape') # Uncomment if needed
            return text
        return None
    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False, # Keep raw param for interface consistency
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Response:
        """Sends a prompt to the Netwrck API and returns the response."""
        if optimizer and optimizer not in self.__available_optimizers:
            raise exceptions.FailedToGenerateResponseError(f"Optimizer is not one of {self.__available_optimizers}")

        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            conversation_prompt = getattr(Optimizers, optimizer)(
                conversation_prompt if conversationally else prompt
            )

        payload = {
            "query": prompt,
            "context": self.system_prompt,
            "examples": [],
            "model_name": self.model_name,
            "greeting": self.greeting
        }

        def for_stream():
            try:
                response = self.session.post(
                    "https://netwrck.com/api/chatpred_or",
                    json=payload,
                    timeout=self.timeout,
                    stream=True,
                    impersonate="chrome110"
                )
                response.raise_for_status()
                buffer = ""
                chunk_size = 32
                for chunk in response.iter_content(chunk_size=chunk_size):
                    if not chunk:
                        continue
                    text = chunk.decode(errors="ignore")
                    buffer += text
                    while len(buffer) >= chunk_size:
                        out = buffer[:chunk_size]
                        buffer = buffer[chunk_size:]
                        if out.strip():
                            if raw:
                                yield out
                            else:
                                yield {"text": out}
                if buffer.strip():
                    if raw:
                        yield buffer
                    else:
                        yield {"text": buffer}
                self.last_response = {"text": buffer}
                self.conversation.update_chat_history(payload["query"], buffer)
            except CurlError as e:
                raise exceptions.APIConnectionError(f"Network error (CurlError): {str(e)}") from e
            except Exception as e:
                err_text = getattr(e, 'response', None) and getattr(e.response, 'text', '')
                raise exceptions.APIConnectionError(f"Unexpected error ({type(e).__name__}): {str(e)} - {err_text}") from e

        def for_non_stream():
            try:
                response = self.session.post(
                    "https://netwrck.com/api/chatpred_or",
                    json=payload,
                    timeout=self.timeout,
                    impersonate="chrome110"
                )
                response.raise_for_status()
                response_text_raw = response.text
                self.last_response = {"text": response_text_raw}
                self.conversation.update_chat_history(prompt, response_text_raw)
                return response_text_raw if raw else self.last_response
            except CurlError as e:
                raise exceptions.FailedToGenerateResponseError(f"Network error (CurlError): {str(e)}") from e
            except Exception as e:
                err_text = getattr(e, 'response', None) and getattr(e.response, 'text', '')
                raise exceptions.FailedToGenerateResponseError(f"Unexpected error ({type(e).__name__}): {str(e)} - {err_text}") from e

        return for_stream() if stream else for_non_stream()

    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
    ) -> Union[str, Generator[str, None, None]]:
        """Generates a response from the Netwrck API."""
        def for_stream_chat():
            # ask() yields dicts or strings when streaming
            gen = self.ask(
                prompt,
                stream=True,
                raw=False, # Ensure ask yields dicts for get_message
                optimizer=optimizer,
                conversationally=conversationally
            )
            for response_dict in gen:
                yield self.get_message(response_dict) # get_message expects dict

        def for_non_stream_chat():
            # ask() returns dict or str when not streaming
            response_data = self.ask(
                prompt,
                stream=False,
                raw=False, # Ensure ask returns dict for get_message
                optimizer=optimizer,
                conversationally=conversationally,
            )
            return self.get_message(response_data) # get_message expects dict

        return for_stream_chat() if stream else for_non_stream_chat()

    def get_message(self, response: Response) -> str:
        """Retrieves message only from response"""
        if not isinstance(response, dict):
            return str(response)
        return response["text"].replace('\\n', '\n').replace('\\n\\n', '\n\n')

if __name__ == "__main__":
    # Ensure curl_cffi is installed
    print("-" * 80)
    print(f"{'Model':<50} {'Status':<10} {'Response'}")
    print("-" * 80)

    # Test all available models
    working = 0
    total = len(Netwrck.AVAILABLE_MODELS)

    for model in Netwrck.AVAILABLE_MODELS:
        try:
            test_ai = Netwrck(model=model, timeout=60)
            response = test_ai.chat("Say 'Hello' in one word", stream=True)
            response_text = ""
            if hasattr(response, "__iter__") and not isinstance(response, (str, bytes)):
                for chunk in response:
                    response_text += chunk
                    print(f"\r{model:<50} {'Testing...':<10}", end="", flush=True)
            else:
                response_text = str(response)

            if response_text and len(response_text.strip()) > 0:
                status = "✓"
                # Truncate response if too long
                display_text = response_text.strip()[:50] + "..." if len(response_text.strip()) > 50 else response_text.strip()
            else:
                status = "✗"
                display_text = "Empty or invalid response"
            print(f"\r{model:<50} {status:<10} {display_text}")
        except Exception as e:
            print(f"\r{model:<50} {'✗':<10} {str(e)}")



from typing import Any, Dict, Generator, Optional, Union

from curl_cffi import CurlError
from curl_cffi.requests import Session

from webscout import exceptions
from webscout.AIbase import Provider, Response
from webscout.AIutel import AwesomePrompts, Conversation, Optimizers, sanitize_stream
from webscout.litagent import LitAgent


class TypefullyAI(Provider):
    required_auth = False
    AVAILABLE_MODELS = [
        "openai:gpt-4o-mini",
        "openai:gpt-4o",
        "anthropic:claude-haiku-4-5-20251001",
        "groq:llama-3.3-70b-versatile",
    ]

    def __init__(
        self,
        is_conversation: bool = True,
        max_tokens: int = 600,
        timeout: int = 30,
        intro: Optional[str] = None,
        filepath: Optional[str] = None,
        update_file: bool = True,
        proxies: dict = {},
        history_offset: int = 10250,
        act: Optional[str] = None,
        system_prompt: str = "You're a helpful assistant.",
        model: str = "openai:gpt-4o-mini",
    ):
        self.session = Session()
        self.is_conversation = is_conversation
        self.max_tokens_to_sample = max_tokens
        self.api_endpoint = "https://typefully.com/tools/ai/api/completion"
        self.timeout = timeout
        self.last_response = {}
        self.system_prompt = system_prompt
        self.model = model
        self.output_length = max_tokens
        self.agent = LitAgent()
        self.headers = {
            "authority": "typefully.com",
            "accept": "*/*",
            "accept-encoding": "gzip, deflate, br, zstd",
            "accept-language": "en-US,en;q=0.9",
            "content-type": "application/json",
            "dnt": "1",
            "origin": "https://typefully.com",
            "referer": "https://typefully.com/tools/ai/chat-gpt-alternative",
            "sec-ch-ua": '"Microsoft Edge";v="135", "Not-A.Brand";v="8", "Chromium";v="135"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Windows"',
            "user-agent": self.agent.random(),
        }
        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )
        self.session.headers.update(self.headers)
        self.session.proxies = proxies
        Conversation.intro = (
            AwesomePrompts().get_act(act, raise_not_found=True, default=None, case_insensitive=True)
            if act
            else intro or Conversation.intro
        )
        self.conversation = Conversation(
            is_conversation, self.max_tokens_to_sample, filepath, update_file
        )
        self.conversation.history_offset = history_offset

    @staticmethod
    def _typefully_extractor(chunk) -> str:
        """Extracts content from Typefully AI SSE format."""
        import json

        # Handle parsed JSON objects (when to_json=True)
        if isinstance(chunk, dict):
            data = chunk
        elif isinstance(chunk, str):
            # Handle raw strings (when to_json=False or direct strings)
            line = chunk.strip()
            if line.startswith("data: "):
                line = line[6:]  # Remove 'data: ' prefix

                try:
                    # Parse JSON content
                    data = json.loads(line)
                except json.JSONDecodeError:
                    # If JSON parsing fails, return empty for non-JSON content
                    return ""
            else:
                # For non-SSE lines, return empty
                return ""
        else:
            # For other types (bytes, etc.), return empty
            return ""

        # Extract delta content for text chunks
        if data.get("type") == "text-delta" and "delta" in data:
            return data["delta"]
        elif data.get("type") == "text-start":
            # Return empty for text-start to avoid duplication
            return ""
        elif data.get("type") == "text-end":
            # Return empty for text-end to avoid duplication
            return ""
        else:
            # Return empty for other event types
            return ""

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Response:
        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            if optimizer in self.__available_optimizers:
                conversation_prompt = getattr(Optimizers, optimizer)(
                    conversation_prompt if conversationally else prompt
                )
            else:
                raise Exception(f"Optimizer is not one of {self.__available_optimizers}")
        payload = {
            "prompt": conversation_prompt,
            "systemPrompt": self.system_prompt,
            "modelIdentifier": self.model,
            "outputLength": self.output_length,
        }

        def for_stream():
            try:
                response = self.session.post(
                    self.api_endpoint,
                    headers=self.headers,
                    json=payload,
                    stream=True,
                    timeout=self.timeout,
                    impersonate="chrome120",
                )
                if not response.ok:
                    raise exceptions.FailedToGenerateResponseError(
                        f"Failed to generate response - ({response.status_code}, {response.reason}) - {response.text}"
                    )
                streaming_text = ""
                processed_stream = sanitize_stream(
                    data=response.iter_content(chunk_size=None),
                    intro_value="data:",
                    to_json=True,
                    content_extractor=self._typefully_extractor,
                    raw=raw,
                )
                for content_chunk in processed_stream:
                    if isinstance(content_chunk, bytes):
                        content_chunk = content_chunk.decode("utf-8", errors="ignore")
                    if content_chunk is None:
                        continue
                    if raw:
                        yield content_chunk
                    else:
                        if content_chunk and isinstance(content_chunk, str):
                            streaming_text += content_chunk
                            yield dict(text=content_chunk)
                self.last_response.update(dict(text=streaming_text))
                self.conversation.update_chat_history(prompt, self.get_message(self.last_response))
            except CurlError as e:
                raise exceptions.FailedToGenerateResponseError(f"Request failed (CurlError): {e}")
            except Exception as e:
                raise exceptions.FailedToGenerateResponseError(
                    f"An unexpected error occurred ({type(e).__name__}): {e}"
                )

        def for_non_stream():
            for _ in for_stream():
                pass
            return self.last_response

        return for_stream() if stream else for_non_stream()

    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        raw: bool = False,
        **kwargs: Any,
    ) -> Union[str, Generator[str, None, None]]:
        def for_stream():
            for response in self.ask(
                prompt, True, raw=raw, optimizer=optimizer, conversationally=conversationally
            ):
                if raw:
                    yield response
                else:
                    yield self.get_message(response)

        def for_non_stream():
            result = self.ask(
                prompt,
                False,
                raw=raw,
                optimizer=optimizer,
                conversationally=conversationally,
            )
            if raw:
                return result
            else:
                return self.get_message(result)

        return for_stream() if stream else for_non_stream()

    def get_message(self, response: Response) -> str:
        if not isinstance(response, dict):
            return str(response)
        text = response.get("text", "")
        try:
            formatted_text = text.replace("\\n", "\n").replace("\\n\\n", "\n\n")
            return formatted_text
        except Exception:
            return text


if __name__ == "__main__":
    print("-" * 80)
    print(f"{'Model':<50} {'Status':<10} {'Response'}")
    print("-" * 80)
    working = 0
    total = len(TypefullyAI.AVAILABLE_MODELS)
    for model in TypefullyAI.AVAILABLE_MODELS:
        try:
            test_ai = TypefullyAI(model=model, timeout=60)
            response_stream = test_ai.chat("Say 'Hello' in one word", stream=True)
            response_text = ""
            if hasattr(response_stream, "__iter__") and not isinstance(response_stream, (str, bytes)):
                for chunk in response_stream:
                    response_text += chunk
            else:
                response_text = str(response_stream)

            if response_text and len(response_text.strip()) > 0:
                status = "OK"
                clean_text = response_text.strip()
                display_text = clean_text[:50] + "..." if len(clean_text) > 50 else clean_text
            else:
                status = "FAIL (Stream)"
                display_text = "Empty or invalid stream response"
            print(f"\r{model:<50} {status:<10} {display_text}")
        except Exception as e:
            print(f"\r{model:<50} {'FAIL':<10} {str(e)}")



import json
import re
import time
import uuid
from typing import Any, Dict, Generator, List, Optional, Union

# Import LitAgent for user agent generation
from webscout.litagent import LitAgent

# Import base classes and utility structures
from webscout.Provider.OPENAI.base import BaseChat, BaseCompletions, OpenAICompatibleProvider
from webscout.Provider.OPENAI.utils import (
    ChatCompletion,
    ChatCompletionChunk,
    ChatCompletionMessage,
    Choice,
    ChoiceDelta,
    CompletionUsage,
    count_tokens,
    format_prompt,
)

# AkashGPT constants
AVAILABLE_MODELS = [
    "Qwen/Qwen3-30B-A3B",
    "DeepSeek-V3.1",
    "Meta-Llama-3-3-70B-Instruct",
]

def _akash_extractor(chunk: Union[str, Dict[str, Any]]) -> Optional[str]:
    """Extracts content from the AkashGPT stream format '0:"..."'."""
    if isinstance(chunk, str):
        match = re.search(r'0:"(.*?)"', chunk)
        if match:
            # Decode potential unicode escapes like \u00e9
            content = match.group(1).encode().decode('unicode_escape')
            return content.replace('\\\\', '\\').replace('\\"', '"') # Handle escaped backslashes and quotes
    return None

class Completions(BaseCompletions):
    def __init__(self, client: 'AkashGPT'):
        self._client = client

    def create(
        self,
        *,
        model: str,
        messages: List[Dict[str, str]],
        max_tokens: Optional[int] = None,
        stream: bool = False,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        timeout: Optional[int] = None,
        proxies: Optional[dict] = None,
        **kwargs: Any
    ) -> Union[ChatCompletion, Generator[ChatCompletionChunk, None, None]]:
        """
        Create a chat completion with AkashGPT API.
        Mimics openai.chat.completions.create
        """
        # Use format_prompt utility to format the conversation
        conversation_prompt = format_prompt(messages, add_special_tokens=True, include_system=True)

        # Set up request parameters
        kwargs.get("api_key", self._client.api_key)

        # Generate request ID and timestamp
        request_id = str(uuid.uuid4())
        created_time = int(time.time())

        # Use the provided model directly
        akash_model = model

        if stream:
            return self._create_streaming(
                request_id, created_time, akash_model, conversation_prompt, messages,
                max_tokens, temperature, top_p, timeout, proxies
            )
        else:
            return self._create_non_streaming(
                request_id, created_time, akash_model, conversation_prompt, messages,
                max_tokens, temperature, top_p, timeout, proxies
            )

    def _create_streaming(
        self, request_id: str, created_time: int, model: str, conversation_prompt: str, messages: List[Dict[str, str]],
        max_tokens: Optional[int], temperature: Optional[float], top_p: Optional[float],
        timeout: Optional[int], proxies: Optional[dict]
    ) -> Generator[ChatCompletionChunk, None, None]:
        """Implementation for streaming chat completions."""
        try:
            # Calculate prompt tokens
            prompt_tokens = count_tokens(conversation_prompt)
            completion_tokens = 0
            total_tokens = 0

            # Make the API request to AkashGPT
            payload = {
                "id": str(uuid.uuid4()).replace("-", ""),
                "messages": messages,
                "model": model,
                "temperature": temperature or 0.6,
                "topP": top_p or 0.9,
                "context": []
            }

            response = self._client.session.post(
                self._client.api_endpoint,
                headers=self._client.headers,
                json=payload,
                stream=True,
                timeout=timeout or 30,
                proxies=proxies
            )

            if not response.ok:
                raise Exception(f"Failed to generate response - ({response.status_code}, {response.reason}) - {response.text}")

            full_content = ""

            # Process the streaming response
            for line in response.iter_lines(decode_unicode=True):
                if line and line.startswith("data: "):
                    data_str = line[6:]
                    if data_str.strip() == "[DONE]":
                        break

                    try:
                        # Try to parse as JSON first
                        data = json.loads(data_str)
                        if isinstance(data, dict) and "text" in data:
                            new_content = data["text"]
                        elif isinstance(data, str):
                            # Use the extractor for raw string responses
                            extracted = _akash_extractor(data_str)
                            new_content = extracted or data_str
                        else:
                            continue

                        if new_content and new_content != full_content:
                            # Calculate delta (new content since last chunk)
                            delta_content = new_content[len(full_content):] if new_content.startswith(full_content) else new_content
                            full_content = new_content
                            completion_tokens = count_tokens(full_content)
                            total_tokens = prompt_tokens + completion_tokens

                            # Only yield chunk if there's new content
                            if delta_content:
                                delta = ChoiceDelta(content=delta_content, role="assistant")
                                choice = Choice(index=0, delta=delta, finish_reason=None)
                                chunk_response = ChatCompletionChunk(
                                    id=request_id,
                                    choices=[choice],
                                    created=created_time,
                                    model=model
                                )
                                chunk_response.usage = {
                                    "prompt_tokens": prompt_tokens,
                                    "completion_tokens": completion_tokens,
                                    "total_tokens": total_tokens
                                }
                                yield chunk_response

                    except json.JSONDecodeError:
                        # Handle non-JSON responses
                        extracted = _akash_extractor(data_str)
                        if extracted and extracted != full_content:
                            delta_content = extracted[len(full_content):] if extracted.startswith(full_content) else extracted
                            full_content = extracted
                            completion_tokens = count_tokens(full_content)
                            total_tokens = prompt_tokens + completion_tokens

                            if delta_content:
                                delta = ChoiceDelta(content=delta_content, role="assistant")
                                choice = Choice(index=0, delta=delta, finish_reason=None)
                                chunk_response = ChatCompletionChunk(
                                    id=request_id,
                                    choices=[choice],
                                    created=created_time,
                                    model=model
                                )
                                chunk_response.usage = {
                                    "prompt_tokens": prompt_tokens,
                                    "completion_tokens": completion_tokens,
                                    "total_tokens": total_tokens
                                }
                                yield chunk_response

            # Final chunk with finish_reason
            delta = ChoiceDelta(content=None)
            choice = Choice(index=0, delta=delta, finish_reason="stop")
            final_chunk = ChatCompletionChunk(
                id=request_id,
                choices=[choice],
                created=created_time,
                model=model
            )
            final_chunk.usage = {
                "prompt_tokens": prompt_tokens,
                "completion_tokens": completion_tokens,
                "total_tokens": total_tokens
            }
            yield final_chunk

        except Exception as e:
            raise IOError(f"AkashGPT streaming request failed: {e}") from e

    def _create_non_streaming(
        self, request_id: str, created_time: int, model: str, conversation_prompt: str, messages: List[Dict[str, str]],
        max_tokens: Optional[int], temperature: Optional[float], top_p: Optional[float],
        timeout: Optional[int], proxies: Optional[dict]
    ) -> ChatCompletion:
        """Implementation for non-streaming chat completions."""
        try:
            # Calculate prompt tokens
            prompt_tokens = count_tokens(conversation_prompt)

            # Make the API request to AkashGPT
            payload = {
                "id": str(uuid.uuid4()).replace("-", ""),
                "messages": messages,
                "model": model,
                "temperature": temperature or 0.6,
                "topP": top_p or 0.9,
                "context": []
            }

            response = self._client.session.post(
                self._client.api_endpoint,
                headers=self._client.headers,
                json=payload,
                timeout=timeout or 30,
                proxies=proxies
            )

            if not response.ok:
                raise Exception(f"Failed to generate response - ({response.status_code}, {response.reason}) - {response.text}")

            # Collect the full response
            full_content = ""
            for line in response.iter_lines(decode_unicode=True):
                if line and line.startswith("data: "):
                    data_str = line[6:]
                    if data_str.strip() == "[DONE]":
                        break

                    try:
                        # Try to parse as JSON first
                        data = json.loads(data_str)
                        if isinstance(data, dict) and "text" in data:
                            full_content = data["text"]
                        elif isinstance(data, str):
                            # Use the extractor for raw string responses
                            extracted = _akash_extractor(data_str)
                            if extracted:
                                full_content = extracted
                    except json.JSONDecodeError:
                        # Handle non-JSON responses
                        extracted = _akash_extractor(data_str)
                        if extracted:
                            full_content = extracted

            # Calculate completion tokens
            completion_tokens = count_tokens(full_content)
            total_tokens = prompt_tokens + completion_tokens

            # Create the completion message
            message = ChatCompletionMessage(
                role="assistant",
                content=full_content
            )

            # Create the choice
            choice = Choice(
                index=0,
                message=message,
                finish_reason="stop"
            )

            usage = CompletionUsage(
                prompt_tokens=prompt_tokens,
                completion_tokens=completion_tokens,
                total_tokens=total_tokens
            )

            # Create the completion object
            completion = ChatCompletion(
                id=request_id,
                choices=[choice],
                created=created_time,
                model=model,
                usage=usage,
            )

            return completion

        except Exception as e:
            raise IOError(f"AkashGPT request failed: {e}") from e


class Chat(BaseChat):
    def __init__(self, client: 'AkashGPT'):
        self.completions = Completions(client)


class AkashGPT(OpenAICompatibleProvider):
    """
    OpenAI-compatible client for AkashGPT API.

    Usage:
        client = AkashGPT(api_key="your_api_key")
        response = client.chat.completions.create(
            model="gpt-4o",
            messages=[{"role": "user", "content": "Hello!"}]
        )
        print(response.choices[0].message.content)
    """
    required_auth = True

    AVAILABLE_MODELS = AVAILABLE_MODELS

    def __init__(
        self,
        api_key: str,
        tools: Optional[List] = None,
        proxies: Optional[Dict[str, str]] = None
    ):
        """
        Initialize the AkashGPT-compatible client.

        Args:
            api_key: Required API key for AkashGPT
            tools: Optional list of tools to register with the provider
            proxies: Optional proxy configuration dict
        """
        super().__init__(api_key=api_key, tools=tools, proxies=proxies)

        # Store the api_key for use in completions
        self.api_key = api_key
        self.timeout = 30
        self.api_endpoint = "https://chat.akash.network/api/chat"

        # Initialize LitAgent for user agent generation
        agent = LitAgent()
        user_agent = agent.random()

        self.headers = {
            "authority": "chat.akash.network",
            "method": "POST",
            "path": "/api/chat",
            "scheme": "https",
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9,en-IN;q=0.8",
            "content-type": "application/json",
            "dnt": "1",
            "origin": "https://chat.akash.network",
            "priority": "u=1, i",
            "referer": "https://chat.akash.network/",
            "sec-ch-ua": '"Microsoft Edge";v="135", "Not-A.Brand";v="8", "Chromium";v="135"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Windows"',
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "sec-gpc": "1",
            "user-agent": user_agent
        }

        self.session.headers.update(self.headers)

        # Initialize chat interface
        self.chat = Chat(self)

    @property
    def models(self):
        class _ModelList:
            def list(inner_self):
                return AVAILABLE_MODELS
        return _ModelList()


if __name__ == "__main__":
    # Example usage
    client = AkashGPT(api_key="your_api_key_here")
    response = client.chat.completions.create(
        model="gpt-4o",
        messages=[{"role": "user", "content": "Hello! How are you?"}]
    )
    print(response.choices[0].message.content)
    print(f"Usage: {response.usage}")



"""Option decorators for SwiftCLI."""

from typing import Any, Callable, Dict, List, Optional, Union

# Sentinel for unspecified defaults
_NO_DEFAULT = object()

def option(
    *param_decls: str,
    type: Any = str,
    required: bool = False,
    default: Any = _NO_DEFAULT,
    help: str = None,
    is_flag: bool = False,
    multiple: bool = False,
    count: bool = False,
    prompt: Union[bool, str] = False,
    hide_input: bool = False,
    confirmation_prompt: Union[bool, str] = False,
    prompt_required: bool = True,
    show_default: bool = True,
    choices: Optional[List[Any]] = None,
    case_sensitive: bool = True,
    callback: Optional[Callable] = None,
    hidden: bool = False,
    validation: Optional[Dict[str, Any]] = None,
    mutually_exclusive: Optional[List[str]] = None
) -> Callable:
    """
    Decorator to add an option to a command.

    Options are named parameters that can be provided in any order.

    Args:
        param_decls: Option names (e.g., "--name", "-n")
        type: Expected type
        required: Whether option is required
        default: Default value
        help: Help text
        is_flag: Whether option is a boolean flag
        multiple: Whether option can be specified multiple times
        count: Whether option counts occurrences
        prompt: Whether to prompt for value if not provided (can be string for prompt text)
        hide_input: Whether to hide input when prompting
        confirmation_prompt: Whether to prompt for confirmation (can be string for prompt text)
        prompt_required: Whether prompt is required
        show_default: Whether to show default in help
        choices: List of valid choices
        case_sensitive: Whether choices are case sensitive
        callback: Function to process/validate value
        hidden: Whether to hide from help output
        validation: Dictionary of validation rules (min_length, max_length, pattern, etc.)
        mutually_exclusive: List of option names that are mutually exclusive with this option

    Example:
        @command()
        @option("--count", "-c", type=int, default=1)
        @option("--format", type=str, choices=["json", "yaml"])
        @option("--verbose", is_flag=True)
        @option("--name", validation={{'min_length': 3, 'max_length': 20}})
        def process(count: int, format: str, verbose: bool, name: str):
            '''Process data'''
            if verbose:
                print(f"Processing {count} items as {format}")
    """
    def decorator(f: Callable) -> Callable:
        if not hasattr(f, '_options'):
            f._options = []

        f._options.append({
            'param_decls': param_decls,
            'type': type,
            'required': required,
            # Only include default explicitly if something other than the sentinel was provided
            **({'default': default} if default is not _NO_DEFAULT else {}),
            'help': help,
            'is_flag': is_flag,
            'multiple': multiple,
            'count': count,
            'prompt': prompt,
            'hide_input': hide_input,
            'confirmation_prompt': confirmation_prompt,
            'prompt_required': prompt_required,
            'show_default': show_default,
            'choices': choices,
            'case_sensitive': case_sensitive,
            'callback': callback,
            'hidden': hidden,
            'validation': validation,
            'mutually_exclusive': mutually_exclusive
        })
        return f
    return decorator

def envvar(
    name: str,
    type: Any = str,
    required: bool = False,
    default: Any = None,
    help: str = None
) -> Callable:
    """
    Decorator to load option value from environment variable.

    Args:
        name: Environment variable name
        type: Expected type
        required: Whether variable is required
        default: Default value if not set
        help: Help text

    Example:
        @command()
        @envvar("API_KEY", required=True)
        @envvar("API_URL", default="https://api.example.com")
        def api_call(api_key: str, api_url: str):
            '''Make API call'''
            print(f"Calling {api_url} with key {api_key}")
    """
    def decorator(f: Callable) -> Callable:
        if not hasattr(f, '_envvars'):
            f._envvars = []

        f._envvars.append({
            'name': name,
            'type': type,
            'required': required,
            'default': default,
            'help': help
        })
        return f
    return decorator

def config_file(
    path: str = None,
    section: str = None,
    required: bool = False,
    auto_create: bool = True,
    format: str = 'json'
) -> Callable:
    """
    Decorator to load configuration from file.

    Args:
        path: Config file path
        section: Config section to load
        required: Whether config is required
        auto_create: Whether to create file if missing
        format: File format (json, yaml, ini)

    Example:
        @command()
        @config_file("~/.myapp/config.json")
        def setup(config: dict):
            '''Setup application'''
            print(f"Database: {config.get('database')}")

        @command()
        @config_file("config.ini", section="api")
        def api(config: dict):
            '''Make API call'''
            print(f"URL: {config.get('url')}")
    """
    def decorator(f: Callable) -> Callable:
        f._config = {
            'path': path,
            'section': section,
            'required': required,
            'auto_create': auto_create,
            'format': format
        }
        return f
    return decorator

def version_option(
    version: str = None,
    prog_name: str = None,
    message: str = None,
    package_name: str = None
) -> Callable:
    """
    Decorator to add version option to command.

    Args:
        version: Version string
        prog_name: Program name
        message: Custom version message
        package_name: Package name to get version from

    Example:
        @command()
        @version_option(version="1.0.0")
        def main():
            '''Main command'''
            pass
    """
    def decorator(f: Callable) -> Callable:
        f._version = {
            'version': version,
            'prog_name': prog_name,
            'message': message,
            'package_name': package_name
        }
        return f
    return decorator

def help_option(
    param_decls: List[str] = None,
    help: str = None
) -> Callable:
    """
    Decorator to customize help option.

    Args:
        param_decls: Help option flags
        help: Help text

    Example:
        @command()
        @help_option(["--help", "-h"], "Show this message")
        def main():
            '''Main command'''
            pass
    """
    def decorator(f: Callable) -> Callable:
        f._help_option = {
            'param_decls': param_decls or ['--help', '-h'],
            'help': help or 'Show this message and exit.'
        }
        return f
    return decorator



##################################################################################
##  ParlerTTS Provider                                                         ##
##################################################################################
import json
import pathlib
import random
import string
import tempfile
from typing import Optional

import httpx
from litprinter import ic

from webscout import exceptions
from webscout.litagent import LitAgent

try:
    from . import utils
    from .base import BaseTTSProvider
except ImportError:
    # Handle direct execution
    import os
    import sys
    sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..', '..'))
    from webscout.Provider.TTS.base import BaseTTSProvider

class ParlerTTS(BaseTTSProvider):
    """
    Text-to-speech provider using the Parler-TTS API (Hugging Face Spaces).

    Features:
    - High-fidelity speech generation
    - Controllable via simple text prompts (description)
    - Manual polling logic for robustness
    """
    required_auth = False

    BASE_URL = "https://parler-tts-parler-tts.hf.space"

    # Request headers
    headers: dict[str, str] = {
        "User-Agent": LitAgent().random(),
        "origin": BASE_URL,
        "referer": f"{BASE_URL}/",
    }

    SUPPORTED_MODELS = ["parler-mini-v1", "parler-large-v1"]

    def __init__(self, timeout: int = 120, proxy: Optional[str] = None):
        """
        Initialize the ParlerTTS client.
        """
        super().__init__()
        self.timeout = timeout
        self.proxy = proxy
        self.default_model = "parler-mini-v1"

    def _generate_session_hash(self) -> str:
        return "".join(random.choices(string.ascii_lowercase + string.digits, k=10))

    def tts(
        self,
        text: str,
        description: str = "A female speaker delivers a slightly expressive and animated speech with a moderate speed. The recording features a low-pitch voice and very clear audio.",
        use_large: bool = False,
        response_format: str = "wav",
        verbose: bool = True
    ) -> str:
        """
        Convert text to speech using Parler-TTS API.
        """
        if not text:
            raise ValueError("Input text must be a non-empty string")

        session_hash = self._generate_session_hash()
        filename = pathlib.Path(tempfile.mktemp(suffix=f".{response_format}", dir=self.temp_dir))

        if verbose:
            ic.configureOutput(prefix='DEBUG| ')
            ic(f"ParlerTTS: Generating speech for '{text[:20]}...'")

        client_kwargs = {"headers": self.headers, "timeout": self.timeout}
        if self.proxy:
            client_kwargs["proxy"] = self.proxy

        try:
            with httpx.Client(**client_kwargs) as client:
                # Step 1: Join the queue
                join_url = f"{self.BASE_URL}/queue/join?__theme=system"
                # fn_index 0 is for the main generation task
                payload = {
                    "data": [text, description, use_large],
                    "event_data": None,
                    "fn_index": 0,
                    "trigger_id": 8,
                    "session_hash": session_hash
                }

                response = client.post(join_url, json=payload)
                response.raise_for_status()

                # Step 2: Poll for data
                data_url = f"{self.BASE_URL}/queue/data?session_hash={session_hash}"
                audio_url = None

                # Gradio Spaces can take time to wake up or process
                with client.stream("GET", data_url) as stream:
                    for line in stream.iter_lines():
                        if not line:
                            continue
                        if line.startswith("data: "):
                            try:
                                data = json.loads(line[6:])
                            except json.JSONDecodeError:
                                continue

                            msg = data.get("msg")
                            if msg == "process_completed":
                                if data.get("success"):
                                    output_data = data.get("output", {}).get("data", [])
                                    if output_data:
                                        audio_info = output_data[0]
                                        path = audio_info["path"] if isinstance(audio_info, dict) else audio_info
                                        audio_url = f"{self.BASE_URL}/file={path}"
                                    break
                                else:
                                    raise exceptions.FailedToGenerateResponseError(f"Generation failed: {data}")
                            elif msg == "queue_full":
                                raise exceptions.FailedToGenerateResponseError("Queue is full")
                            elif msg == "send_hash":
                                # Normal handshake
                                pass

                if not audio_url:
                    raise exceptions.FailedToGenerateResponseError("Failed to get audio URL from stream")

                # Step 3: Download the audio file
                audio_response = client.get(audio_url)
                audio_response.raise_for_status()

                with open(filename, "wb") as f:
                    f.write(audio_response.content)

                if verbose:
                    ic.configureOutput(prefix='DEBUG| ')
                    ic(f"Speech generated successfully: {filename}")

                return filename.as_posix()

        except Exception as e:
            if verbose:
                ic.configureOutput(prefix='DEBUG| ')
                ic(f"Error in ParlerTTS: {e}")
            raise exceptions.FailedToGenerateResponseError(f"Failed to generate audio: {e}")

    def create_speech(self, input: str, **kwargs) -> str:
        return self.tts(text=input, **kwargs)

if __name__ == "__main__":
    tts = ParlerTTS()
    try:
        path = tts.tts("Testing Parler-TTS with manual polling.", verbose=True)
        ic.configureOutput(prefix='INFO| ')
        ic(f"Saved to {path}")
    except Exception as e:
        ic.configureOutput(prefix='ERROR| ')
        ic(f"Error: {e}")



import re
import time
import uuid
from typing import Any, Dict, Generator, List, Optional, Union

from curl_cffi.requests import Session

from webscout.AIutel import sanitize_stream
from webscout.Provider.OPENAI.base import BaseChat, BaseCompletions, OpenAICompatibleProvider
from webscout.Provider.OPENAI.utils import (
    ChatCompletion,
    ChatCompletionChunk,
    ChatCompletionMessage,
    Choice,
    ChoiceDelta,
    CompletionUsage,
)

try:
    from webscout.litagent import LitAgent
except ImportError:
    pass

class Completions(BaseCompletions):
    def __init__(self, client: 'Elmo'):
        self._client = client

    @staticmethod
    def _elmo_extractor(chunk: Union[str, Dict[str, Any]]) -> Optional[str]:
        """Extracts content from the Elmo stream format '0:"..."'."""
        if isinstance(chunk, str):
            match = re.search(r'0:"(.*?)"(?=,|$)', chunk) # Look for 0:"...", possibly followed by comma or end of string
            if match:
                # Decode potential unicode escapes like \u00e9 and handle escaped quotes/backslashes
                content = match.group(1).encode().decode('unicode_escape')
                return content.replace('\\\\', '\\').replace('\\"', '"')
        return None

    def create(
        self,
        *,
        model: str,
        messages: List[Dict[str, str]],
        max_tokens: Optional[int] = 600,
        stream: bool = False,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        timeout: Optional[int] = None,
        proxies: Optional[Dict[str, str]] = None,
        **kwargs: Any
    ) -> Union[ChatCompletion, Generator[ChatCompletionChunk, None, None]]:
        # Elmo uses a fixed system prompt, but we can override if provided
        system_prompt = self._client.system_prompt
        for msg in messages:
            if msg.get('role') == 'system':
                system_prompt = msg.get('content', system_prompt)
                break

        # Build conversation in Elmo format
        conversation = []
        if system_prompt:
            conversation.append({"role": "system", "content": system_prompt})
        for msg in messages:
            if msg['role'] in ['user', 'assistant']:
                conversation.append({"role": msg['role'], "content": msg['content']})

        payload = {
            "metadata": {
                "system": {"language": "en-US"},
                "website": {
                    "url": "chrome-extension://ipnlcfhfdicbfbchfoihipknbaeenenm/options.html",
                    "origin": "chrome-extension://ipnlcfhfdicbfbchfoihipknbaeenenm",
                    "title": "Elmo Chat - Your AI Web Copilot",
                    "xpathIndexLength": 0,
                    "favicons": [],
                    "language": "en",
                    "content": "",
                    "type": "html",
                    "selection": "",
                    "hash": "d41d8cd98f00b204e9800998ecf8427e",
                },
            },
            "regenerate": True,
            "conversation": conversation,
            "enableCache": False,
        }

        request_id = f"chatcmpl-{uuid.uuid4()}"
        created_time = int(time.time())
        if stream:
            return self._create_stream(request_id, created_time, model, payload, timeout, proxies)
        else:
            return self._create_non_stream(request_id, created_time, model, payload, timeout, proxies)

    def _create_stream(
        self, request_id: str, created_time: int, model: str, payload: Dict[str, Any],
        timeout: Optional[int] = None, proxies: Optional[Dict[str, str]] = None
    ) -> Generator[ChatCompletionChunk, None, None]:
        try:
            response = self._client.session.post(
                self._client.api_endpoint,
                headers=self._client.headers,
                json=payload,
                stream=True,
                timeout=timeout or self._client.timeout,
                proxies=proxies,
                impersonate="chrome110"
            )
            response.raise_for_status()

            # Use sanitize_stream to process the response
            processed_stream = sanitize_stream(
                data=response.iter_content(chunk_size=None), # Pass byte iterator
                intro_value=None, # No simple prefix
                to_json=False,    # Content is text after extraction
                content_extractor=self._elmo_extractor, # Use the specific extractor
                yield_raw_on_error=True,
                raw=False
            )

            prompt_tokens = 0
            completion_tokens = 0
            total_tokens = 0

            for content_chunk in processed_stream:
                if content_chunk and isinstance(content_chunk, str):
                    completion_tokens += len(content_chunk.split())
                    total_tokens = prompt_tokens + completion_tokens
                    delta = ChoiceDelta(
                        content=content_chunk,
                        role="assistant",
                        tool_calls=None
                    )
                    choice = Choice(
                        index=0,
                        delta=delta,
                        finish_reason=None,
                        logprobs=None
                    )
                    chunk = ChatCompletionChunk(
                        id=request_id,
                        choices=[choice],
                        created=created_time,
                        model=model,
                        system_fingerprint=None
                    )
                    chunk.usage = {
                        "prompt_tokens": prompt_tokens,
                        "completion_tokens": completion_tokens,
                        "total_tokens": total_tokens,
                        "estimated_cost": None
                    }
                    yield chunk

            # Final chunk
            delta = ChoiceDelta(content=None, role=None, tool_calls=None)
            choice = Choice(index=0, delta=delta, finish_reason="stop", logprobs=None)
            chunk = ChatCompletionChunk(
                id=request_id,
                choices=[choice],
                created=created_time,
                model=model,
                system_fingerprint=None
            )
            chunk.usage = {
                "prompt_tokens": prompt_tokens,
                "completion_tokens": completion_tokens,
                "total_tokens": total_tokens,
                "estimated_cost": None
            }
            yield chunk
        except Exception as e:
            print(f"Error during Elmo stream request: {e}")
            raise IOError(f"Elmo request failed: {e}") from e

    def _create_non_stream(
        self, request_id: str, created_time: int, model: str, payload: Dict[str, Any],
        timeout: Optional[int] = None, proxies: Optional[Dict[str, str]] = None
    ) -> ChatCompletion:
        try:
            response = self._client.session.post(
                self._client.api_endpoint,
                headers=self._client.headers,
                json=payload,
                stream=True,  # Still use stream=True to get the response as a stream
                timeout=timeout or self._client.timeout,
                proxies=proxies
            )
            response.raise_for_status()

            # Use sanitize_stream to process the response and aggregate content
            processed_stream = sanitize_stream(
                data=response.iter_content(chunk_size=None), # Pass byte iterator
                intro_value=None, # No simple prefix
                to_json=False,    # Content is text after extraction
                content_extractor=self._elmo_extractor, # Use the specific extractor
                yield_raw_on_error=True,
                raw=False
            )

            # Aggregate all content
            content = ""
            for content_chunk in processed_stream:
                if content_chunk and isinstance(content_chunk, str):
                    content += content_chunk

            message = ChatCompletionMessage(
                role="assistant",
                content=content
            )
            choice = Choice(
                index=0,
                message=message,
                finish_reason="stop"
            )
            usage = CompletionUsage(
                prompt_tokens=0,  # Elmo doesn't provide token counts
                completion_tokens=len(content.split()),
                total_tokens=len(content.split())
            )
            completion = ChatCompletion(
                id=request_id,
                choices=[choice],
                created=created_time,
                model=model,
                usage=usage,
            )
            return completion
        except Exception as e:
            print(f"Error during Elmo non-stream request: {e}")
            raise IOError(f"Elmo request failed: {e}") from e

class Chat(BaseChat):
    def __init__(self, client: 'Elmo'):
        self.completions = Completions(client)

class Elmo(OpenAICompatibleProvider):
    required_auth = False
    AVAILABLE_MODELS = ["elmo"]

    def __init__(self, browser: str = "chrome"):
        self.timeout = 30
        self.api_endpoint = "https://www.elmo.chat/api/v1/prompt"
        self.system_prompt = "You are a helpful AI assistant. Provide clear, concise, and well-structured information. Organize your responses into paragraphs for better readability."
        self.session = Session()
        self.headers = {
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9,en-IN;q=0.8",
            "content-type": "text/plain;charset=UTF-8",
            "dnt": "1",
            "origin": "chrome-extension://ipnlcfhfdicbfbchfoihipknbaeenenm",
            "priority": "u=1, i",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "cross-site",
        }
        self.session.headers.update(self.headers)
        self.chat = Chat(self)

    @property
    def models(self):
        class _ModelList:
            def list(inner_self):
                return type(self).AVAILABLE_MODELS
        return _ModelList()

if __name__ == "__main__":
    client = Elmo()
    response = client.chat.completions.create(
        model="elmo",
        messages=[{"role": "user", "content": "Hello, how are you?"}],
        max_tokens=600,
        stream=False
    )
    print(response.choices[0].message.content)



"""
Webscout OpenAI-Compatible API Server

A FastAPI-based server that provides OpenAI-compatible endpoints for various LLM providers.
Supports streaming and non-streaming chat completions with comprehensive error handling,
authentication, and provider management.
"""

from __future__ import annotations

import os

import uvicorn
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.openapi.docs import get_swagger_ui_html
from starlette.responses import HTMLResponse

from .config import AppConfig, ServerConfig
from .providers import initialize_provider_map, initialize_tti_provider_map
from .routes import Api
from .ui_templates import LANDING_PAGE_HTML, SWAGGER_CSS

# Configuration constants
DEFAULT_PORT = 8000
DEFAULT_HOST = "0.0.0.0"
API_VERSION = "v1"

# Setup logging
# Using litprinter console instead of standard logging

# Global configuration instance - lazy initialization
config = None

def get_config() -> ServerConfig:
    """Get or create the global configuration instance."""
    global config
    if config is None:
        config = ServerConfig()
    return config


def create_app():
    """Create and configure the FastAPI application."""
    app_title = os.getenv("WEBSCOUT_API_TITLE", "Webscout API")
    app_description = os.getenv("WEBSCOUT_API_DESCRIPTION", "OpenAI API compatible interface for various LLM providers")
    app_version = os.getenv("WEBSCOUT_API_VERSION", "0.2.0")
    app_docs_url = os.getenv("WEBSCOUT_API_DOCS_URL", "/docs")
    app_redoc_url = os.getenv("WEBSCOUT_API_REDOC_URL", "/redoc")
    app_openapi_url = os.getenv("WEBSCOUT_API_OPENAPI_URL", "/openapi.json")

    app = FastAPI(
        title=app_title,
        description=app_description,
        version=app_version,
        docs_url=None,  # Disable default docs
        redoc_url=app_redoc_url,
        openapi_url=app_openapi_url,
    )

    # Simple Custom Swagger UI with WebScout footer
    @app.get(app_docs_url, include_in_schema=False)
    async def custom_swagger_ui_html():
        html = get_swagger_ui_html(
            openapi_url=app.openapi_url,
            title=app.title + " - API Documentation",
        ).body.decode("utf-8")

        # Custom footer and styles
        footer_html = """
        <div class="webscout-footer">
            Powered by <a href='https://github.com/OEvortex/Webscout' target='_blank'>WebScout</a>
        </div>
        """

        # Inject custom CSS and footer
        html = html.replace("</head>", f"<style>{SWAGGER_CSS}</style></head>")
        html = html.replace("</body>", f"{footer_html}</body>")
        return HTMLResponse(content=html)

    # Add CORS middleware
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )


    # Add startup event handler
    @app.on_event("startup")
    async def startup():
        if hasattr(app.state, 'startup_event'):
            await app.state.startup_event()

    # Initialize API routes
    api = Api(app)
    api.register_validation_exception_handler()
    api.register_routes()

    # Initialize providers
    initialize_provider_map()
    initialize_tti_provider_map()

    # Root landing page
    @app.get("/", include_in_schema=False)
    async def root():
        return HTMLResponse(content=LANDING_PAGE_HTML)

    return app


def create_app_debug():
    """Create app in debug mode."""
    return create_app()


def start_server(
    port: int = DEFAULT_PORT,
    host: str = DEFAULT_HOST,
    default_provider: str = None,
    base_url: str = None,
    workers: int = 1,
    log_level: str = 'info',
    debug: bool = False,
):
    """Start the API server with the given configuration."""
    run_api(
        host=host,
        port=port,
        default_provider=default_provider,
        base_url=base_url,
        workers=workers,
        log_level=log_level,
        debug=debug,
    )


def run_api(
    host: str = '0.0.0.0',
    port: int = None,
    default_provider: str = None,
    base_url: str = None,
    debug: bool = False,
    workers: int = 1,
    log_level: str = 'info',
    show_available_providers: bool = True,
) -> None:
    """Run the API server with configuration."""
    print("Starting Webscout OpenAI API server...")
    if port is None:
        port = DEFAULT_PORT

    AppConfig.set_config(
        api_key=None,
        default_provider=default_provider or AppConfig.default_provider,
        base_url=base_url,
        auth_required=False,
        rate_limit_enabled=False
    )

    if show_available_providers:
        if not AppConfig.provider_map:
            initialize_provider_map()
        if not AppConfig.tti_provider_map:
            initialize_tti_provider_map()

        print("\n=== Webscout OpenAI API Server ===")
        print(f"Server URL: http://{host if host != '0.0.0.0' else 'localhost'}:{port}")
        if AppConfig.base_url:
            print(f"Base Path: {AppConfig.base_url}")
            api_endpoint_base = f"http://{host if host != '0.0.0.0' else 'localhost'}:{port}{AppConfig.base_url}"
        else:
            api_endpoint_base = f"http://{host if host != '0.0.0.0' else 'localhost'}:{port}"

        print(f"API Endpoint: {api_endpoint_base}/v1/chat/completions")
        print(f"Docs URL: {api_endpoint_base}/docs")

        # Show authentication status
        print("Authentication: 🔓 DISABLED")

        # Show rate limiting status
        print("Rate Limiting: ⚡ DISABLED")

        print(f"Default Provider: {AppConfig.default_provider}")
        print(f"Workers: {workers}")
        print(f"Log Level: {log_level}")
        print(f"Debug Mode: {'Enabled' if debug else 'Disabled'}")

        providers = list(set(v.__name__ for v in AppConfig.provider_map.values()))
        print(f"\n--- Available Providers ({len(providers)}) ---")
        for i, provider_name in enumerate(sorted(providers), 1):
            print(f"{i}. {provider_name}")

        provider_class_names = set(v.__name__ for v in AppConfig.provider_map.values())
        models = sorted([model for model in AppConfig.provider_map.keys() if model not in provider_class_names])
        if models:
            print(f"\n--- Available Models ({len(models)}) ---")
            for i, model_name in enumerate(models, 1):
                print(f"{i}. {model_name} (via {AppConfig.provider_map[model_name].__name__})")
        else:
            print("\nNo specific models registered. Use provider names as models.")

        tti_providers = list(set(v.__name__ for v in AppConfig.tti_provider_map.values()))
        print(f"\n--- Available TTI Providers ({len(tti_providers)}) ---")
        for i, provider_name in enumerate(sorted(tti_providers), 1):
            print(f"{i}. {provider_name}")

        tti_models = sorted([model for model in AppConfig.tti_provider_map.keys() if model not in tti_providers])
        if tti_models:
            print(f"\n--- Available TTI Models ({len(tti_models)}) ---")
            for i, model_name in enumerate(tti_models, 1):
                print(f"{i}. {model_name} (via {AppConfig.tti_provider_map[model_name].__name__})")
        else:
            print("\nNo specific TTI models registered. Use TTI provider names as models.")

        print("\nUse Ctrl+C to stop the server.")
        print("=" * 40 + "\n")

    uvicorn_app_str = "webscout.server.server:create_app_debug" if debug else "webscout.server.server:create_app"

    # Configure uvicorn settings
    uvicorn_config = {
        "app": uvicorn_app_str,
        "host": host,
        "port": int(port),
        "factory": True,
        "reload": debug,
        "log_level": log_level.lower() if log_level else ("debug" if debug else "info"),
    }

    # Add workers only if not in debug mode
    if not debug and workers > 1:
        uvicorn_config["workers"] = workers
        print(f"Starting with {workers} workers...")
    elif debug:
        print("Debug mode enabled - using single worker with reload...")

    uvicorn.run(**uvicorn_config)


def main():
    """Main entry point for the webscout-server console script."""
    import argparse

    # Read environment variables with fallbacks
    default_port = int(os.getenv('WEBSCOUT_PORT', os.getenv('PORT', DEFAULT_PORT)))
    default_host = os.getenv('WEBSCOUT_HOST', DEFAULT_HOST)
    default_workers = int(os.getenv('WEBSCOUT_WORKERS', '1'))
    default_log_level = os.getenv('WEBSCOUT_LOG_LEVEL', 'info')
    default_provider = os.getenv('WEBSCOUT_DEFAULT_PROVIDER', os.getenv('DEFAULT_PROVIDER'))
    default_base_url = os.getenv('WEBSCOUT_BASE_URL', os.getenv('BASE_URL'))
    default_debug = os.getenv('WEBSCOUT_DEBUG', os.getenv('DEBUG', 'false')).lower() == 'true'

    parser = argparse.ArgumentParser(description='Start Webscout OpenAI-compatible API server')
    parser.add_argument('--port', type=int, default=default_port, help=f'Port to run the server on (default: {default_port})')
    parser.add_argument('--host', type=str, default=default_host, help=f'Host to bind the server to (default: {default_host})')
    parser.add_argument('--workers', type=int, default=default_workers, help=f'Number of worker processes (default: {default_workers})')
    parser.add_argument('--log-level', type=str, default=default_log_level, choices=['debug', 'info', 'warning', 'error', 'critical'], help=f'Log level (default: {default_log_level})')
    parser.add_argument('--default-provider', type=str, default=default_provider, help='Default provider to use (optional)')
    parser.add_argument('--base-url', type=str, default=default_base_url, help='Base URL for the API (optional, e.g., /api/v1)')
    parser.add_argument('--debug', action='store_true', default=default_debug, help='Run in debug mode')
    args = parser.parse_args()

    # Print configuration summary
    print("Configuration:")
    print(f"  Host: {args.host}")
    print(f"  Port: {args.port}")
    print(f"  Workers: {args.workers}")
    print(f"  Log Level: {args.log_level}")
    print(f"  Debug Mode: {args.debug}")
    print("  Authentication: 🔓 DISABLED")
    print("  Rate Limiting: ⚡ DISABLED")
    print(f"  Default Provider: {args.default_provider or 'Not set'}")
    print(f"  Base URL: {args.base_url or 'Not set'}")
    print()

    run_api(
        host=args.host,
        port=args.port,
        workers=args.workers,
        log_level=args.log_level,
        default_provider=args.default_provider,
        base_url=args.base_url,
        debug=args.debug
    )


if __name__ == "__main__":
    main()



import json
from typing import Any, Dict, Generator, Optional, Union

from curl_cffi.requests import Session

from webscout import exceptions
from webscout.AIbase import Provider
from webscout.AIutel import AwesomePrompts, Conversation, Optimizers, sanitize_stream


class samurai(Provider):
    """
    A class to interact with a custom API endpoint.
    """
    AVAILABLE_MODELS = [
        "openai/gpt-4.1",
        "openai/gpt-4o-search-preview",
        "openai/gpt-4o-mini-search-preview",
        "openai/gpt-4.1-nano",
        "openai/chatgpt-4o-latest",
        "openai/gpt-4.1-mini",
        "gpt-4o",
        "o3-mini",
        "Claude-sonnet-3.7",
        "uncensored-r1",
        "anthropic/claude-3.5-sonnet",
        "gemini-1.5-pro",
        "gemini-1.5-pro-latest",
        "gemini-flash-2.0",
        "gemini-1.5-flash",
        "llama-3.1-405b",
        "Meta-Llama-3.1-405B-Instruct-Turbo",
        "Meta-Llama-3.3-70B-Instruct-Turbo",
        "chutesai/Llama-4-Maverick-17B-128E-Instruct-FP8",
        "chutesai/Llama-4-Scout-17B-16E-Instruct",
        "Qwen-QwQ-32B-Preview",
        "Qwen/Qwen3-235B-A22B-fp8-tput",
        "deepseek-r1",
        "deepseek-v3",
        "deepseek-ai/DeepSeek-V3-0324",
        "dbrx-instruct",
        "x-ai/grok-3-",
        "perplexity-ai/r1-1776"
    ]
    def __init__(
        self,
        is_conversation: bool = True,
        max_tokens: int = 2049,
        timeout: int = 30,
        intro: str = None,
        filepath: str = None,
        update_file: bool = True,
        proxies: dict = {},
        history_offset: int = 10250,
        act: str = None,
        model: str = "openai/gpt-4.1",
        system_prompt: str = "You are a helpful assistant."
    ):
        """Initializes the Custom API client."""
        self.url = "https://newapi-9qln.onrender.com/v1/chat/completions"
        self.headers = {
            "Authorization": "Bearer Samurai-AP1-Fr33",
            "Content-Type": "application/json"
        }
        self.session = Session()
        self.session.headers.update(self.headers)
        self.system_prompt = system_prompt
        self.is_conversation = is_conversation
        self.max_tokens_to_sample = max_tokens
        self.timeout = timeout
        self.last_response = {}
        self.model = model

        self.__available_optimizers = (
            method
            for method in dir(Optimizers)
            if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )
        Conversation.intro = (
            AwesomePrompts().get_act(
                act, raise_not_found=True, default=None, case_insensitive=True
            )
            if act
            else intro or Conversation.intro
        )

        self.conversation = Conversation(
            is_conversation, self.max_tokens_to_sample, filepath, update_file
        )
        self.conversation.history_offset = history_offset

    @staticmethod
    def _extractor(chunk: dict) -> Optional[str]:
        """Extracts content from stream JSON objects."""
        return chunk.get("choices", [{}])[0].get("delta", {}).get("content")

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
    ) -> Union[Dict[str, Any], Generator]:
        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            if optimizer in self.__available_optimizers:
                conversation_prompt = getattr(Optimizers, optimizer)(
                    conversation_prompt if conversationally else prompt
                )
            else:
                raise Exception(f"Optimizer is not one of {self.__available_optimizers}")

        payload = {
            "model": self.model,
            "messages": [
                {"role": "system", "content": self.system_prompt},
                {"role": "user", "content": conversation_prompt},
            ],
            "stream": stream
        }

        def for_stream():
            streaming_text = ""
            try:
                response = self.session.post(
                    self.url,
                    data=json.dumps(payload),
                    stream=True,
                    timeout=self.timeout,
                    impersonate="chrome110"
                )
                response.raise_for_status()

                processed_stream = sanitize_stream(
                    data=response.iter_lines(),
                    intro_value="data:",
                    to_json=True,
                    skip_markers=["[DONE]"],
                    content_extractor=self._extractor,
                    yield_raw_on_error=False
                )

                for content_chunk in processed_stream:
                    if content_chunk and isinstance(content_chunk, str):
                        streaming_text += content_chunk
                        resp = dict(text=content_chunk)
                        yield resp if not raw else content_chunk

            except Exception as e:
                raise exceptions.FailedToGenerateResponseError(f"Request failed: {str(e)}") from e
            finally:
                if streaming_text:
                    self.last_response = {"text": streaming_text}
                    self.conversation.update_chat_history(prompt, streaming_text)

        def for_non_stream():
            try:
                response = self.session.post(
                self.url,
                data=json.dumps(payload),
                timeout=self.timeout,
                impersonate="chrome110"
            )
                response.raise_for_status()
                json_response = response.json()
                content = json_response.get("choices", [{}])[0].get("message", {}).get("content", "")

                self.last_response = {"text": content}
                self.conversation.update_chat_history(prompt, content)
                return self.last_response if not raw else content

            except Exception as e:
                raise exceptions.FailedToGenerateResponseError(f"Request failed: {str(e)}") from e

        return for_stream() if stream else for_non_stream()

    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: str = None,
        conversationally: bool = False,
    ) -> Union[str, Generator[str, None, None]]:
        def for_stream_chat():
            gen = self.ask(prompt, stream=True, raw=False, optimizer=optimizer, conversationally=conversationally)
            for response_dict in gen:
                yield self.get_message(response_dict)

        def for_non_stream_chat():
            response_data = self.ask(prompt, stream=False, raw=False, optimizer=optimizer, conversationally=conversationally)
            return self.get_message(response_data)

        return for_stream_chat() if stream else for_non_stream_chat()

    def get_message(self, response: dict) -> str:
        assert isinstance(response, dict), "Response should be of dict data-type only"
        return response["text"]

if __name__ == "__main__":
    # Ensure curl_cffi is installed
    print("-" * 80)
    print(f"{'Model':<50} {'Status':<10} {'Response'}")
    print("-" * 80)

    for model in samurai.AVAILABLE_MODELS:
        try:
            test_ai = samurai(model=model, timeout=60)
            response = test_ai.chat("Say 'Hello' in one word", stream=True)
            response_text = ""
            for chunk in response:
                response_text += chunk

            if response_text and len(response_text.strip()) > 0:
                status = "✓"
                # Clean and truncate response
                clean_text = response_text.strip().encode('utf-8', errors='ignore').decode('utf-8')
                display_text = clean_text[:50] + "..." if len(clean_text) > 50 else clean_text
            else:
                status = "✗"
                display_text = "Empty or invalid response"
            print(f"\r{model:<50} {status:<10} {display_text}")
        except Exception as e:
            print(f"\r{model:<50} {'✗':<10} {str(e)}")



import re
from typing import Any, Dict, Generator, List, Optional, Union

import requests

from webscout import exceptions
from webscout.AIbase import AISearch, SearchResponse
from webscout.litagent import LitAgent
from webscout.sanitize import sanitize_stream


class webpilotai(AISearch):
    """A class to interact with the webpilotai (WebPilot) AI search API.

    webpilotai provides a web-based comprehensive search SearchResponse interface that returns AI-generated
    SearchResponses with source references and related questions. It supports both streaming and
    non-streaming SearchResponses.

    Basic Usage:
        >>> from webscout import webpilotai
        >>> ai = webpilotai()
        >>> # Non-streaming example
        >>> response = ai.search("What is Python?")
        >>> print(response)
        Python is a high-level programming language...

        >>> # Streaming example
        >>> for chunk in ai.search("Tell me about AI", stream=True):
        ...     print(chunk, end="", flush=True)
        Artificial Intelligence is...

        >>> # Raw SearchResponse format
        >>> for chunk in ai.search("Hello", stream=True, raw=True):
        ...     print(chunk)
        {'text': 'Hello'}
        {'text': ' there!'}

    Args:
        timeout (int, optional): Request timeout in seconds. Defaults to 90.
        proxies (dict, optional): Proxy configuration for requests. Defaults to None.
    """

    def __init__(
        self,
        timeout: int = 90,
        proxies: Optional[dict] = None,
    ):
        """Initialize the webpilotai API client.

        Args:
            timeout (int, optional): Request timeout in seconds. Defaults to 90.
            proxies (dict, optional): Proxy configuration for requests. Defaults to None.
        """
        self.session = requests.Session()
        self.api_endpoint = "https://api.webpilotai.com/rupee/v1/search"
        self.timeout = timeout
        self.last_response = {}

        # The 'Bearer null' is part of the API's expected headers
        self.headers = {
            'Accept': 'application/json, text/plain, */*, text/event-stream',
            'Content-Type': 'application/json;charset=UTF-8',
            'Authorization': 'Bearer null',
            'Origin': 'https://www.webpilot.ai',
            'Referer': 'https://www.webpilot.ai/',
            'User-Agent': LitAgent().random(),
        }

        self.session.headers.update(self.headers)
        self.proxies = proxies

    def search(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
    ) -> Union[SearchResponse, Generator[Union[Dict[str, str], SearchResponse], None, None], List[Any]]:
        """Search using the webpilotai API and get AI-generated SearchResponses.

        This method sends a search query to webpilotai and returns the AI-generated SearchResponse.
        It supports both streaming and non-streaming modes, as well as raw SearchResponse format.

        Args:
            prompt (str): The search query or prompt to send to the API.
            stream (bool, optional): If True, yields SearchResponse chunks as they arrive.
                                   If False, returns complete SearchResponse. Defaults to False.
            raw (bool, optional): If True, returns raw SearchResponse dictionaries with 'text' key.
                                If False, returns SearchResponse objects that convert to text automatically.
                                Defaults to False.

        Returns:
            Union[SearchResponse, Generator[Union[Dict[str, str], SearchResponse], None, None]]:
                - If stream=False: Returns complete SearchResponse as SearchResponse object
                - If stream=True: Yields SearchResponse chunks as either Dict or SearchResponse objects

        Raises:
            APIConnectionError: If the API request fails
        """
        payload = {
            "q": prompt,
            "threadId": ""  # Empty for new search
        }

        def for_stream():

            try:
                with self.session.post(
                    self.api_endpoint,
                    json=payload,
                    stream=True,
                    timeout=self.timeout,
                    proxies=self.proxies
                ) as response:
                    if not response.ok:
                        raise exceptions.APIConnectionError(
                            f"Failed to generate response - ({response.status_code}, {response.reason}) - {response.text}"
                        )

                    def extract_webpilot_content(data):
                        if isinstance(data, dict):
                            if data.get('type') == 'data':
                                return data.get('data', {}).get('content', "")
                        return None

                    processed_chunks = sanitize_stream(
                        data=response.iter_content(chunk_size=1024),
                        to_json=True,
                        extract_regexes=[r"(\{.*\})"],
                        content_extractor=lambda chunk: extract_webpilot_content(chunk),
                        yield_raw_on_error=False,
                        encoding='utf-8',
                        encoding_errors='replace',
                        line_delimiter="\n",
                        raw=raw,
                        output_formatter=None if raw else lambda x: SearchResponse(x) if isinstance(x, str) else x,
                    )

                    for chunk in processed_chunks:
                        yield chunk

            except requests.exceptions.Timeout:
                raise exceptions.APIConnectionError("Request timed out")
            except requests.exceptions.RequestException as e:
                raise exceptions.APIConnectionError(f"Request failed: {e}")

        def for_non_stream():
            full_content = ""
            for chunk in for_stream():
                if raw:
                    full_content += str(chunk)
                else:
                    full_content += str(chunk)

            if raw:
                return full_content
            else:
                formatted_response = self.format_SearchResponse(full_content)
                self.last_response = SearchResponse(formatted_response)
                return self.last_response

        if stream:
            return for_stream()
        else:
            return for_non_stream()

    @staticmethod
    def format_SearchResponse(text: str) -> str:
        """Format the SearchResponse text for better readability.

        Args:
            text (str): The raw SearchResponse text

        Returns:
            str: Formatted text with improved structure
        """
        # Clean up formatting
        # Remove excessive newlines
        clean_text = re.sub(r'\n{3,}', '\n\n', text)

        # Ensure consistent spacing around sections
        clean_text = re.sub(r'([.!?])\s*\n\s*([A-Z])', r'\1\n\n\2', clean_text)

        # Clean up any leftover HTML or markdown artifacts
        clean_text = re.sub(r'<[^>]*>', '', clean_text)

        # Remove trailing whitespace on each line
        clean_text = '\n'.join(line.rstrip() for line in clean_text.split('\n'))

        return clean_text.strip()


if __name__ == "__main__":
    from rich import print

    ai = webpilotai()
    r = ai.search("What is Python?", stream=True, raw=False)
    for chunk in r:
        print(chunk, end="", flush=True)



##################################################################################
##  Deepgram TTS Provider                                                      ##
##################################################################################
import pathlib
import tempfile
import time
from concurrent.futures import ThreadPoolExecutor, as_completed

import requests
from litprinter import ic

from webscout import exceptions
from webscout.litagent import LitAgent

try:
    from . import utils
    from .base import BaseTTSProvider
except ImportError:
    # Handle direct execution
    import os
    import sys
    sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..', '..', '..'))
    from webscout.Provider.TTS import utils
    from webscout.Provider.TTS.base import BaseTTSProvider

class DeepgramTTS(BaseTTSProvider):
    """
    Text-to-speech provider using the Deepgram Aura-2 API.

    This provider follows the OpenAI TTS API structure with support for:
    - Aura-2 next-gen voices
    - Low-latency real-time performance
    - Multiple output formats
    - Concurrent generation for long texts
    """
    required_auth = False

    # Request headers
    headers: dict[str, str] = {
        "Accept": "*/*",
        "Accept-Language": "en-US,en;q=0.9",
        "Content-Type": "application/json",
        "Origin": "https://deepgram.com",
        "Referer": "https://deepgram.com/ai-voice-generator",
        "User-Agent": LitAgent().random()
    }

    # Supported Aura-2 voices
    SUPPORTED_MODELS = ["aura-2"]

    SUPPORTED_VOICES = [
        "thalia", "odysseus", "harmonia", "theia", "electra",
        "arcas", "amalthea", "helena", "hyperion", "apollo", "luna",
        # Legacy Aura-1 voices (if still supported by the endpoint)
        "asteria", "luna", "stella", "athena", "hera",
        "zeus", "orpheus", "arcas", "perseus", "angus", "orion", "helios"
    ]

    # Voice mapping for Deepgram API compatibility
    voice_mapping = {
        # Aura-2
        "thalia": "aura-2-thalia-en",
        "odysseus": "aura-2-odysseus-en",
        "harmonia": "aura-2-harmonia-en",
        "theia": "aura-2-theia-en",
        "electra": "aura-2-electra-en",
        "arcas": "aura-2-arcas-en",
        "amalthea": "aura-2-amalthea-en",
        "helena": "aura-2-helena-en",
        "hyperion": "aura-2-hyperion-en",
        "apollo": "aura-2-apollo-en",
        "luna": "aura-2-luna-en",
        # Aura-1 (Backward compatibility)
        "asteria": "aura-asteria-en",
        "stella": "aura-stella-en",
        "athena": "aura-athena-en",
        "hera": "aura-hera-en",
        "zeus": "aura-zeus-en",
        "orpheus": "aura-orpheus-en",
        "perseus": "aura-perseus-en",
        "angus": "aura-angus-en",
        "orion": "aura-orion-en",
        "helios": "aura-helios-en"
    }

    def __init__(self, timeout: int = 30, proxies: dict = None):
        """
        Initialize the Deepgram TTS client.

        Args:
            timeout (int): Request timeout in seconds
            proxies (dict): Proxy configuration
        """
        super().__init__()
        self.api_url = "https://deepgram.com/api/ttsAudioGeneration"
        self.session = requests.Session()
        self.session.headers.update(self.headers)
        if proxies:
            self.session.proxies.update(proxies)
        self.timeout = timeout
        self.default_voice = "thalia"

    def tts(
        self,
        text: str,
        model: str = "aura-2", # Dummy model param for compatibility
        voice: str = "thalia",
        response_format: str = "mp3",
        instructions: str = None,
        verbose: bool = True
    ) -> str:
        """
        Convert text to speech using Deepgram Aura-2 API.

        Args:
            text (str): The text to convert to speech
            voice (str): The voice to use (thalia, odysseus, etc.)
            response_format (str): Audio format (mp3, wav, aac, flac, opus, pcm)
            verbose (bool): Whether to print debug information

        Returns:
            str: Path to the generated audio file
        """
        if not text:
            raise ValueError("Input text must be a non-empty string")

        # Map voice to Deepgram API format
        voice_id = self.voice_mapping.get(voice.lower(), f"aura-2-{voice.lower()}-en")

        # Create temporary file
        file_extension = f".{response_format}"
        filename = pathlib.Path(tempfile.mktemp(suffix=file_extension, dir=self.temp_dir))

        # Split text into sentences for long inputs
        sentences = utils.split_sentences(text)
        if verbose:
            ic.configureOutput(prefix='DEBUG| ')
            ic(f"DeepgramTTS: Processing {len(sentences)} chunks")
            ic.configureOutput(prefix='DEBUG| ')
            ic(f"Voice: {voice} -> {voice_id}")

        def generate_audio_for_chunk(part_text: str, part_number: int):
            max_retries = 3
            for attempt in range(max_retries):
                try:
                    payload = {
                        "text": part_text,
                        "model": voice_id,
                        "demoType": "voice-generator",
                        "params": "tag=landingpage-aivoicegenerator"
                    }
                    response = self.session.post(
                        self.api_url,
                        json=payload,
                        timeout=self.timeout
                    )
                    response.raise_for_status()

                    if response.content:
                        if verbose:
                            ic.configureOutput(prefix='DEBUG| ')
                            ic(f"Chunk {part_number} processed successfully")
                        return part_number, response.content

                except requests.RequestException as e:
                    if verbose:
                        ic.configureOutput(prefix='WARNING| ')
                        ic(f"Error processing chunk {part_number}: {e}. Retrying {attempt+1}/{max_retries}")
                    time.sleep(1)

            raise exceptions.FailedToGenerateResponseError(f"Failed to generate audio for chunk {part_number}")

        try:
            with ThreadPoolExecutor(max_workers=3) as executor:
                futures = {
                    executor.submit(generate_audio_for_chunk, sentence.strip(), i): i
                    for i, sentence in enumerate(sentences)
                }

                audio_chunks = {}
                for future in as_completed(futures):
                    part_num, data = future.result()
                    audio_chunks[part_num] = data

                with open(filename, 'wb') as f:
                    for i in sorted(audio_chunks.keys()):
                        f.write(audio_chunks[i])

                if verbose:
                    ic.configureOutput(prefix='INFO| ')
                    ic(f"Audio saved to {filename}")

                return str(filename)

        except Exception as e:
            raise exceptions.FailedToGenerateResponseError(f"Deepgram TTS failed: {e}")

    def create_speech(self, input: str, **kwargs) -> str:
        """OpenAI-compatible speech creation interface."""
        return self.tts(text=input, **kwargs)

    def with_streaming_response(self):
        return StreamingResponseContextManager(self)

class StreamingResponseContextManager:
    def __init__(self, tts_provider: DeepgramTTS):
        self.tts_provider = tts_provider
    def create(self, **kwargs):
        audio_file = self.tts_provider.create_speech(**kwargs)
        return StreamingResponse(audio_file)
    def __enter__(self): return self
    def __exit__(self, exc_type, exc_val, exc_tb): pass

class StreamingResponse:
    def __init__(self, audio_file: str):
        self.audio_file = audio_file
    def __enter__(self): return self
    def __exit__(self, exc_type, exc_val, exc_tb): pass
    def stream_to_file(self, file_path: str):
        import shutil
        shutil.copy2(self.audio_file, file_path)
    def iter_bytes(self, chunk_size: int = 1024):
        with open(self.audio_file, 'rb') as f:
            while chunk := f.read(chunk_size):
                yield chunk

if __name__ == "__main__":
    dg = DeepgramTTS()
    try:
        path = dg.tts("Deepgram Aura-2 test successful. Generating high quality speech.", voice="thalia", verbose=True)
        print(f"Saved to: {path}")
    except Exception as e:
        print(f"Error: {e}")



import json
import re
from typing import Any, Dict, Generator, List, Optional

from .https import video_data

try:
    from webscout.litagent.agent import LitAgent
    _USER_AGENT = LitAgent().random()
except ImportError:
    _USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36"


class Video:

    _HEAD = 'https://www.youtube.com/watch?v='

    def __init__(self, video_id: str):
        """
        Represents a YouTube video

        Parameters
        ----------
        video_id : str
            The id or url of the video
        """
        pattern = re.compile(r'.be/(.*?)$|=(.*?)$|shorts/(.*?)$|^(\w{11})$')  # noqa
        match = pattern.search(video_id)

        if not match:
            raise ValueError('Invalid YouTube video ID or URL')

        self._matched_id = (
                match.group(1)
                or match.group(2)
                or match.group(3)
                or match.group(4)
        )

        if self._matched_id:
            self._url = self._HEAD + self._matched_id
            self._video_data = video_data(self._matched_id)
            # Extract basic info for fallback
            title_match = re.search(r'<title>(.*?) - YouTube</title>', self._video_data)
            self.title = title_match.group(1) if title_match else None
            self.id = self._matched_id
        else:
            raise ValueError('Invalid YouTube video ID or URL')

    def __repr__(self):
        return f'<Video {self._url}>'

    @property
    def metadata(self) -> Dict[str, Any]:
        """
        Fetches video metadata in a dict format

        Returns
        -------
        Dict
            Video metadata in a dict format containing keys: title, id, views, duration, author_id,
            upload_date, url, thumbnails, tags, description, likes, genre, etc.
        """
        # Multiple patterns to try for video details extraction for robustness
        details_patterns = [
            re.compile(r'videoDetails\":(.*?)\"isLiveContent\":.*?}'),
            re.compile(r'videoDetails\":(.*?),\"playerConfig'),
            re.compile(r'videoDetails\":(.*?),\"playabilityStatus')
        ]

        # Other metadata patterns
        upload_date_pattern = re.compile(r"<meta itemprop=\"uploadDate\" content=\"(.*?)\">")
        genre_pattern = re.compile(r"<meta itemprop=\"genre\" content=\"(.*?)\">")
        like_count_patterns = [
            re.compile(r"iconType\":\"LIKE\"},\"defaultText\":(.*?)}"),
            re.compile(r'\"likeCount\":\"(\d+)\"')
        ]
        channel_name_pattern = re.compile(r'"ownerChannelName":"(.*?)"')

        # Try each pattern for video details
        raw_details_match = None
        for pattern in details_patterns:
            match = pattern.search(self._video_data)
            if match:
                raw_details_match = match
                break

        if not raw_details_match:
            # Fallback metadata for search results or incomplete video data
            return {
                'title': getattr(self, 'title', None),
                'id': getattr(self, 'id', None),
                'views': getattr(self, 'views', None),
                'streamed': False,
                'duration': None,
                'author_id': None,
                'author_name': None,
                'upload_date': None,
                'url': f"https://www.youtube.com/watch?v={getattr(self, 'id', '')}" if hasattr(self, 'id') else None,
                'thumbnails': None,
                'tags': None,
                'description': None,
                'likes': None,
                'genre': None,
                'is_age_restricted': 'age-restricted' in self._video_data.lower(),
                'is_unlisted': 'unlisted' in self._video_data.lower()
            }

        raw_details = raw_details_match.group(0)

        # Extract upload date
        upload_date_match = upload_date_pattern.search(self._video_data)
        upload_date = upload_date_match.group(1) if upload_date_match else None

        # Extract channel name
        channel_name_match = channel_name_pattern.search(self._video_data)
        channel_name = channel_name_match.group(1) if channel_name_match else None

        # Parse video details
        try:
            # Clean up the JSON string for parsing
            clean_json = raw_details.replace('videoDetails\":', '')
            # Handle potential JSON parsing issues
            if clean_json.endswith(','):
                clean_json = clean_json[:-1]
            metadata = json.loads(clean_json)

            data = {
                'title': metadata.get('title'),
                'id': metadata.get('videoId', self._matched_id),
                'views': metadata.get('viewCount'),
                'streamed': metadata.get('isLiveContent', False),
                'duration': metadata.get('lengthSeconds'),
                'author_id': metadata.get('channelId'),
                'author_name': channel_name or metadata.get('author'),
                'upload_date': upload_date,
                'url': f"https://www.youtube.com/watch?v={metadata.get('videoId', self._matched_id)}",
                'thumbnails': metadata.get('thumbnail', {}).get('thumbnails'),
                'tags': metadata.get('keywords'),
                'description': metadata.get('shortDescription'),
                'is_age_restricted': metadata.get('isAgeRestricted', False) or 'age-restricted' in self._video_data.lower(),
                'is_unlisted': 'unlisted' in self._video_data.lower(),
                'is_family_safe': metadata.get('isFamilySafe', True),
                'is_private': metadata.get('isPrivate', False),
                'is_live_content': metadata.get('isLiveContent', False),
                'is_crawlable': metadata.get('isCrawlable', True),
                'allow_ratings': metadata.get('allowRatings', True)
            }
        except (json.JSONDecodeError, KeyError, TypeError) as e:
            # Fallback to basic metadata if JSON parsing fails
            return {
                'title': getattr(self, 'title', None),
                'id': self._matched_id,
                'url': self._url,
                'error': f"Failed to parse video details: {str(e)}"
            }

        # Try to extract likes count
        likes = None
        for pattern in like_count_patterns:
            try:
                likes_match = pattern.search(self._video_data)
                if likes_match:
                    likes_text = likes_match.group(1)
                    # Handle different formats of like count
                    if '{' in likes_text:
                        likes = json.loads(likes_text + '}}}')['accessibility']['accessibilityData']['label'].split(' ')[0].replace(',', '')
                    else:
                        likes = likes_text
                    break
            except (AttributeError, KeyError, json.decoder.JSONDecodeError):
                continue

        data['likes'] = likes

        # Try to extract genre
        try:
            genre_match = genre_pattern.search(self._video_data)
            data['genre'] = genre_match.group(1) if genre_match else None
        except AttributeError:
            data['genre'] = None

        return data

    @property
    def embed_html(self) -> str:
        """
        Get the embed HTML code for this video

        Returns:
            HTML iframe code for embedding the video
        """
        return f'<iframe width="560" height="315" src="https://www.youtube.com/embed/{self._matched_id}" frameborder="0" allow="accelerometer; autoplay; clipboard-write; encrypted-media; gyroscope; picture-in-picture" allowfullscreen></iframe>'

    @property
    def embed_url(self) -> str:
        """
        Get the embed URL for this video

        Returns:
            URL for embedding the video
        """
        return f'https://www.youtube.com/embed/{self._matched_id}'

    @property
    def thumbnail_url(self) -> str:
        """
        Get the thumbnail URL for this video

        Returns:
            URL of the video thumbnail (high quality)
        """
        return f'https://i.ytimg.com/vi/{self._matched_id}/hqdefault.jpg'

    @property
    def thumbnail_urls(self) -> Dict[str, str]:
        """
        Get all thumbnail URLs for this video in different qualities

        Returns:
            Dictionary of thumbnail URLs with quality labels
        """
        return {
            'default': f'https://i.ytimg.com/vi/{self._matched_id}/default.jpg',
            'medium': f'https://i.ytimg.com/vi/{self._matched_id}/mqdefault.jpg',
            'high': f'https://i.ytimg.com/vi/{self._matched_id}/hqdefault.jpg',
            'standard': f'https://i.ytimg.com/vi/{self._matched_id}/sddefault.jpg',
            'maxres': f'https://i.ytimg.com/vi/{self._matched_id}/maxresdefault.jpg'
        }

    @property
    def is_live(self) -> bool:
        """
        Check if video is currently live.

        Returns:
            True if video is live, False otherwise
        """
        return '"isLive":true' in self._video_data or '"isLiveNow":true' in self._video_data

    @property
    def is_short(self) -> bool:
        """
        Check if video is a YouTube Short.

        Returns:
            True if video is a Short, False otherwise
        """
        # Check duration (Shorts are max 60 seconds)
        meta = self.metadata
        duration = meta.get('duration')
        if duration and int(duration) <= 60:
            # Also check for shorts indicators
            if 'shorts' in self._video_data.lower() or '"shorts"' in self._video_data:
                return True
        return False

    @property
    def hashtags(self) -> List[str]:
        """
        Get hashtags from video description and title.

        Returns:
            List of hashtags found
        """
        meta = self.metadata
        text = (meta.get('description', '') or '') + ' ' + (meta.get('title', '') or '')
        pattern = r'#([a-zA-Z0-9_]+)'
        matches = re.findall(pattern, text)
        return list(dict.fromkeys(matches))  # Remove duplicates

    def get_related_videos(self, limit: int = 10) -> List[str]:
        """
        Get related/suggested videos.

        Args:
            limit: Maximum number of video IDs to return

        Returns:
            List of related video IDs
        """
        # Find related videos in the page data
        pattern = r'"watchNextEndScreenRenderer".*?"videoId":"([a-zA-Z0-9_-]{11})"'
        matches = re.findall(pattern, self._video_data)

        if not matches:
            # Fallback pattern
            pattern = r'"compactVideoRenderer".*?"videoId":"([a-zA-Z0-9_-]{11})"'
            matches = re.findall(pattern, self._video_data)

        # Remove duplicates and self
        seen = set()
        unique = []
        for vid in matches:
            if vid not in seen and vid != self._matched_id:
                seen.add(vid)
                unique.append(vid)
                if len(unique) >= limit:
                    break

        return unique

    def get_chapters(self) -> Optional[List[Dict[str, Any]]]:
        """
        Get video chapters if available.

        Returns:
            List of chapters with title, start_time, and thumbnail, or None
        """
        # Look for chapter data in the page
        pattern = r'"chapterRenderer":\s*\{[^}]*"title":\s*\{\s*"simpleText":\s*"([^"]+)"[^}]*"timeRangeStartMillis":\s*(\d+)'
        matches = re.findall(pattern, self._video_data)

        if not matches:
            # Alternative pattern
            pattern = r'"chapters":\s*\[(.*?)\]'
            chapter_match = re.search(pattern, self._video_data)
            if chapter_match:
                try:
                    # Parse as JSON if possible
                    chapters_str = '[' + chapter_match.group(1) + ']'
                    chapters_data = json.loads(chapters_str)
                    return chapters_data
                except Exception:
                    pass
            return None

        chapters = []
        for title, start_ms in matches:
            chapters.append({
                'title': title,
                'start_seconds': int(start_ms) / 1000,
                'start_time': f"{int(int(start_ms)/1000//60)}:{int(int(start_ms)/1000%60):02d}"
            })

        return chapters if chapters else None

    def stream_comments(self, limit: int = 20) -> Generator[Dict[str, Any], None, None]:
        """
        Stream video comments from initial page load.

        Note: YouTube loads comments dynamically via JavaScript, so this method
        may not find comments for all videos. It works best when YouTube includes
        some initial comments in the page HTML.

        Args:
            limit: Maximum number of comments to yield

        Yields:
            Comment dictionaries with author, text, video_id
        """
        # Try multiple patterns for comments data
        patterns = [
            # Pattern for commentRenderer with simpleText
            r'"commentRenderer":\s*\{[^}]*"authorText":\s*\{[^}]*"simpleText":\s*"([^"]+)"[^}]*\}.*?"contentText":\s*\{[^}]*"runs":\s*\[\s*\{[^}]*"text":\s*"([^"]*)"',
            # Pattern for author and contentText
            r'"authorText":\s*\{[^}]*"simpleText":\s*"([^"]+)"[^}]*\}[^}]*"contentText":\s*\{[^}]*"text":\s*"([^"]*)"',
            # Simpler pattern
            r'"authorText":"([^"]+)".*?"contentText":"([^"]*)"',
        ]

        count = 0
        seen_comments = set()

        for pattern in patterns:
            if count >= limit:
                break
            matches = re.findall(pattern, self._video_data, re.DOTALL)
            for author, text in matches:
                if count >= limit:
                    break
                # Avoid duplicates
                comment_key = (author, text[:50])
                if comment_key in seen_comments:
                    continue
                seen_comments.add(comment_key)

                # Clean up text
                text = text.replace('\\n', '\n')
                text = text.replace('\\u0026', '&')
                text = text.replace('\\u003c', '<')
                text = text.replace('\\u003e', '>')

                yield {
                    'author': author,
                    'text': text,
                    'video_id': self._matched_id
                }
                count += 1


if __name__ == '__main__':
    video = Video('https://www.youtube.com/watch?v=9bZkp7q19f0')
    print(video.metadata)
    print(f"\nIs Live: {video.is_live}")
    print(f"Is Short: {video.is_short}")
    print(f"Hashtags: {video.hashtags}")
    print(f"Related videos: {video.get_related_videos(5)}")

    chapters = video.get_chapters()
    if chapters:
        print(f"Chapters: {chapters[:3]}")




"""
Gradient Network OpenAI-compatible Provider
Reverse engineered from https://chat.gradient.network/

Provides OpenAI-compatible API interface for Gradient Network's distributed GPU clusters.
"""

import json
import time
import uuid
from typing import Any, Dict, Generator, List, Optional, Union

import requests

from webscout.Provider.OPENAI.base import BaseChat, BaseCompletions, OpenAICompatibleProvider
from webscout.Provider.OPENAI.utils import (
    ChatCompletion,
    ChatCompletionChunk,
    ChatCompletionMessage,
    Choice,
    ChoiceDelta,
    CompletionUsage,
    count_tokens,
)

# ANSI escape codes for formatting
BOLD = "\033[1m"
RED = "\033[91m"
RESET = "\033[0m"


class Completions(BaseCompletions):
    """Handles chat completions for Gradient Network."""

    def __init__(self, client: 'Gradient'):
        self._client = client

    def create(
        self,
        *,
        model: str,
        messages: List[Dict[str, str]],
        max_tokens: Optional[int] = 2048,
        stream: bool = False,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        timeout: Optional[int] = None,
        proxies: Optional[Dict[str, str]] = None,
        enable_thinking: Optional[bool] = None,
        cluster_mode: Optional[str] = None,
        **kwargs: Any
    ) -> Union[ChatCompletion, Generator[ChatCompletionChunk, None, None]]:
        """
        Creates a model response for the given chat conversation.
        Mimics openai.chat.completions.create

        Args:
            model: The model to use for completion
            messages: List of message dicts with 'role' and 'content'
            max_tokens: Maximum tokens to generate (not used by Gradient API)
            stream: Whether to stream the response
            temperature: Sampling temperature (not used by Gradient API)
            top_p: Top-p sampling (not used by Gradient API)
            timeout: Request timeout in seconds
            proxies: Proxy configuration
            enable_thinking: Enable thinking/reasoning mode (default: client setting)
            cluster_mode: GPU cluster mode (auto-detected based on model if None)
            **kwargs: Additional arguments

        Returns:
            ChatCompletion or Generator[ChatCompletionChunk] depending on stream
        """
        # Convert model name and get appropriate cluster mode
        converted_model = self._client.convert_model_name(model)
        actual_cluster_mode = cluster_mode or self._client.MODEL_CLUSTERS.get(converted_model, self._client.cluster_mode)

        # Build the payload - pass messages directly as the API accepts them
        payload = {
            "model": converted_model,
            "clusterMode": actual_cluster_mode,
            "messages": messages,
            "enableThinking": enable_thinking if enable_thinking is not None else self._client.enable_thinking,
        }

        request_id = f"chatcmpl-{uuid.uuid4()}"
        created_time = int(time.time())

        if stream:
            return self._create_stream(request_id, created_time, model, payload, timeout, proxies)
        else:
            return self._create_non_stream(request_id, created_time, model, payload, timeout, proxies)

    def _create_stream(
        self, request_id: str, created_time: int, model: str, payload: Dict[str, Any],
        timeout: Optional[int] = None, proxies: Optional[Dict[str, str]] = None
    ) -> Generator[ChatCompletionChunk, None, None]:
        """Handle streaming response from Gradient API."""
        try:
            response = self._client.session.post(
                self._client.base_url,
                headers=self._client.headers,
                json=payload,
                stream=True,
                timeout=timeout or self._client.timeout,
                proxies=proxies or self._client.proxies or None
            )
            response.raise_for_status()

            completion_tokens = 0
            prompt_tokens = count_tokens(str(payload.get("messages", [])))
            first_chunk = True

            for line in response.iter_lines():
                if not line:
                    continue

                try:
                    # Decode bytes to string
                    decoded_line = line.decode('utf-8') if isinstance(line, bytes) else line
                    decoded_line = decoded_line.strip()
                    if not decoded_line:
                        continue

                    # Parse JSON response
                    data = json.loads(decoded_line)

                    # Only process "reply" type chunks
                    chunk_type = data.get("type")
                    if chunk_type != "reply":
                        continue

                    # Extract content - prefer "content" over "reasoningContent"
                    reply_data = data.get("data", {})
                    content = reply_data.get("content") or reply_data.get("reasoningContent")

                    if content:
                        completion_tokens += count_tokens(content)

                        delta = ChoiceDelta(
                            content=content,
                            role="assistant" if first_chunk else None
                        )
                        first_chunk = False

                        choice = Choice(
                            index=0,
                            delta=delta,
                            finish_reason=None,
                            logprobs=None
                        )

                        chunk = ChatCompletionChunk(
                            id=request_id,
                            choices=[choice],
                            created=created_time,
                            model=model,
                            system_fingerprint=None
                        )
                        chunk.usage = {
                            "prompt_tokens": prompt_tokens,
                            "completion_tokens": completion_tokens,
                            "total_tokens": prompt_tokens + completion_tokens,
                            "estimated_cost": None
                        }
                        yield chunk

                except json.JSONDecodeError:
                    continue
                except Exception:
                    continue

            # Final chunk with finish_reason="stop"
            delta = ChoiceDelta(content=None, role=None, tool_calls=None)
            choice = Choice(index=0, delta=delta, finish_reason="stop", logprobs=None)
            chunk = ChatCompletionChunk(
                id=request_id,
                choices=[choice],
                created=created_time,
                model=model,
                system_fingerprint=None
            )
            chunk.usage = {
                "prompt_tokens": prompt_tokens,
                "completion_tokens": completion_tokens,
                "total_tokens": prompt_tokens + completion_tokens,
                "estimated_cost": None
            }
            yield chunk

        except requests.exceptions.RequestException as e:
            print(f"{RED}Error during Gradient stream request: {e}{RESET}")
            raise IOError(f"Gradient request failed: {e}") from e
        except Exception as e:
            print(f"{RED}Error during Gradient stream request: {type(e).__name__}: {e}{RESET}")
            raise IOError(f"Gradient request failed: {e}") from e

    def _create_non_stream(
        self, request_id: str, created_time: int, model: str, payload: Dict[str, Any],
        timeout: Optional[int] = None, proxies: Optional[Dict[str, str]] = None
    ) -> ChatCompletion:
        """Handle non-streaming response from Gradient API."""
        try:
            # Collect all chunks from streaming
            full_content = ""
            prompt_tokens = count_tokens(str(payload.get("messages", [])))

            response = self._client.session.post(
                self._client.base_url,
                headers=self._client.headers,
                json=payload,
                stream=True,
                timeout=timeout or self._client.timeout,
                proxies=proxies or self._client.proxies or None
            )
            response.raise_for_status()

            for line in response.iter_lines():
                if not line:
                    continue

                try:
                    # Decode bytes to string
                    decoded_line = line.decode('utf-8') if isinstance(line, bytes) else line
                    decoded_line = decoded_line.strip()
                    if not decoded_line:
                        continue

                    data = json.loads(decoded_line)

                    # Only process "reply" type chunks
                    chunk_type = data.get("type")
                    if chunk_type != "reply":
                        continue

                    reply_data = data.get("data", {})
                    # Prefer "content" over "reasoningContent"
                    content = reply_data.get("content") or reply_data.get("reasoningContent")
                    if content:
                        full_content += content

                except json.JSONDecodeError:
                    continue
                except Exception:
                    continue

            completion_tokens = count_tokens(full_content)
            total_tokens = prompt_tokens + completion_tokens

            usage = CompletionUsage(
                prompt_tokens=prompt_tokens,
                completion_tokens=completion_tokens,
                total_tokens=total_tokens
            )

            message = ChatCompletionMessage(
                role="assistant",
                content=full_content
            )

            choice = Choice(
                index=0,
                message=message,
                finish_reason="stop"
            )

            completion = ChatCompletion(
                id=request_id,
                choices=[choice],
                created=created_time,
                model=model,
                usage=usage,
            )

            return completion

        except requests.exceptions.RequestException as e:
            print(f"{RED}Error during Gradient non-stream request: {e}{RESET}")
            raise IOError(f"Gradient request failed: {e}") from e
        except Exception as e:
            print(f"{RED}Error during Gradient non-stream request: {type(e).__name__}: {e}{RESET}")
            raise IOError(f"Gradient request failed: {e}") from e


class Chat(BaseChat):
    """Chat interface for Gradient Network."""

    def __init__(self, client: 'Gradient'):
        self.completions = Completions(client)


class Gradient(OpenAICompatibleProvider):
    """
    OpenAI-compatible client for Gradient Network API.

    Gradient Network provides access to distributed GPU clusters running large language models.
    This provider supports real-time streaming responses.

    Note: GPT OSS 120B works on "nvidia" cluster, Qwen3 235B works on "hybrid" cluster.
    Cluster mode is auto-detected based on model selection.

    Usage:
        client = Gradient()
        response = client.chat.completions.create(
            model="GPT OSS 120B",
            messages=[{"role": "user", "content": "Hello!"}]
        )
        print(response.choices[0].message.content)

        # Streaming
        for chunk in client.chat.completions.create(
            model="Qwen3 235B",
            messages=[{"role": "user", "content": "Tell me a story"}],
            stream=True
        ):
            if chunk.choices[0].delta.content:
                print(chunk.choices[0].delta.content, end="", flush=True)
    """

    required_auth = False

    AVAILABLE_MODELS = [
        "GPT OSS 120B",
        "Qwen3 235B",
    ]

    # Model to cluster mapping
    MODEL_CLUSTERS = {
        "GPT OSS 120B": "nvidia",
        "Qwen3 235B": "hybrid",
    }

    def __init__(
        self,
        timeout: int = 60,
        cluster_mode: str = "nvidia",
        enable_thinking: bool = True,
        proxies: Optional[Dict[str, str]] = None,
    ):
        """
        Initialize the Gradient client.

        Args:
            timeout: Request timeout in seconds (default: 60)
            cluster_mode: Default GPU cluster mode (default: "nvidia", auto-detected per model)
            enable_thinking: Enable thinking/reasoning mode (default: True)
            proxies: Optional proxy configuration
        """
        self.timeout = timeout
        self.cluster_mode = cluster_mode
        self.enable_thinking = enable_thinking
        self.proxies = proxies or {}

        self.base_url = "https://chat.gradient.network/api/generate"
        self.session = requests.Session()

        # Set up headers matching the curl request
        self.headers = {
            "accept": "*/*",
            "accept-language": "en-US,en;q=0.9,en-IN;q=0.8",
            "content-type": "application/json",
            "dnt": "1",
            "origin": "https://chat.gradient.network",
            "priority": "u=1, i",
            "referer": "https://chat.gradient.network/",
            "sec-ch-ua": '"Microsoft Edge";v="143", "Chromium";v="143", "Not A(Brand";v="24"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"Windows"',
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "sec-gpc": "1",
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/143.0.0.0 Safari/537.36 Edg/143.0.0.0",
        }

        self.session.headers.update(self.headers)

        # Initialize the chat interface
        self.chat = Chat(self)

    def convert_model_name(self, model: str) -> str:
        """
        Ensure the model name is in the correct format.

        Args:
            model: Model name to convert

        Returns:
            Valid model name
        """
        if model in self.AVAILABLE_MODELS:
            return model

        # Try case-insensitive matching with dash to space conversion
        model_lower = model.lower().replace("-", " ")
        for available_model in self.AVAILABLE_MODELS:
            if model_lower == available_model.lower():
                return available_model

        # Default to first available model
        print(f"{BOLD}Warning: Model '{model}' not found, using default model '{self.AVAILABLE_MODELS[0]}'{RESET}")
        return self.AVAILABLE_MODELS[0]

    @property
    def models(self):
        """Returns a model list interface."""
        class _ModelList:
            def list(inner_self):
                return type(self).AVAILABLE_MODELS
        return _ModelList()


if __name__ == "__main__":
    print("-" * 80)
    print(f"{'Model':<50} {'Status':<10} {'Response'}")
    print("-" * 80)

    for model in Gradient.AVAILABLE_MODELS:
        try:
            client = Gradient(timeout=120)
            response = client.chat.completions.create(
                model=model,
                messages=[
                    {"role": "user", "content": "Say 'Hello' in one word"},
                ],
                stream=False
            )

            if response and response.choices and response.choices[0].message.content:
                status = "✓"
                display_text = response.choices[0].message.content.strip()
                display_text = display_text[:50] + "..." if len(display_text) > 50 else display_text
            else:
                status = "✗"
                display_text = "Empty or invalid response"
            print(f"{model:<50} {status:<10} {display_text}")
        except Exception as e:
            print(f"{model:<50} {'✗':<10} {str(e)[:50]}")



import requests


def create_grammar_check_job(text: str):
    url = 'https://api.aigrammarchecker.io/api/ai-check-grammar/create-job'
    headers = {
        'accept': '*/*',
        'accept-language': 'en-US,en;q=0.9,en-IN;q=0.8',
        'dnt': '1',
        'origin': 'https://aigrammarchecker.io',
        'priority': 'u=1, i',
        'product-code': '067003',
        'product-serial': '6a4836a29e756bd24a74ebed31e405da',
        'referer': 'https://aigrammarchecker.io/',
        'sec-ch-ua': '"Not(A:Brand";v="8", "Chromium";v="120", "Microsoft Edge";v="120"',
        'sec-ch-ua-mobile': '?0',
        'sec-ch-ua-platform': '"Windows"',
        'sec-fetch-dest': 'empty',
        'sec-fetch-mode': 'cors',
        'sec-fetch-site': 'same-site',
        'sec-gpc': '1',
        'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36 Edg/120.0.0.0',
    }

    files = {
        'features': (None, 'check_grammar'),
        'entertext': (None, text),
        'translate_language': (None, 'English'),
    }

    response = requests.post(url, headers=headers, files=files)
    return response.json()

if __name__ == "__main__":
    from rich import print as cprint
    cprint(create_grammar_check_job("she gg to school"))



from webscout.Provider.AISEARCH.genspark_search import Genspark
from webscout.Provider.AISEARCH.iask_search import IAsk
from webscout.Provider.AISEARCH.monica_search import Monica
from webscout.Provider.AISEARCH.PERPLEXED_search import PERPLEXED
from webscout.Provider.AISEARCH.Perplexity import Perplexity
from webscout.Provider.AISEARCH.webpilotai_search import webpilotai

# List of all exported names
__all__ = [
    "PERPLEXED",
    "Perplexity",
    "Genspark",
    "IAsk",
    "Monica",
    "webpilotai",
]



"""
Base class for TTS providers with OpenAI-compatible functionality.
"""
import os
import tempfile
from pathlib import Path
from typing import Generator

from litprinter import ic

from webscout.AIbase import TTSProvider


class BaseTTSProvider(TTSProvider):
    """
    Base class for TTS providers with OpenAI-compatible functionality.

    This class implements common methods and follows OpenAI TTS API patterns
    for speech generation, streaming, and audio handling.
    """
    required_auth = False

    # Supported models (can be overridden by subclasses)
    SUPPORTED_MODELS = [
        "gpt-4o-mini-tts",  # Latest intelligent realtime model
        "tts-1",            # Lower latency model
        "tts-1-hd"          # Higher quality model
    ]

    # Supported voices (can be overridden by subclasses)
    SUPPORTED_VOICES = [
        "alloy", "ash", "ballad", "coral", "echo",
        "fable", "nova", "onyx", "sage", "shimmer"
    ]

    # Supported output formats
    SUPPORTED_FORMATS = [
        "mp3",    # Default format
        "opus",   # Internet streaming, low latency
        "aac",    # Digital audio compression
        "flac",   # Lossless compression
        "wav",    # Uncompressed, low latency
        "pcm"     # Raw samples, 24kHz 16-bit
    ]

    def __init__(self):
        """Initialize the base TTS provider."""
        self.temp_dir = tempfile.mkdtemp(prefix="webscout_tts_")
        self.default_model = "gpt-4o-mini-tts"
        self.default_voice = "coral"
        self.default_format = "mp3"

    def validate_model(self, model: str) -> str:
        """
        Validate and return the model name.

        Args:
            model (str): Model name to validate

        Returns:
            str: Validated model name

        Raises:
            ValueError: If model is not supported
        """
        # If provider doesn't support models, return the model as-is
        if self.SUPPORTED_MODELS is None:
            return model

        if model not in self.SUPPORTED_MODELS:
            raise ValueError(f"Model '{model}' not supported. Available models: {', '.join(self.SUPPORTED_MODELS)}")
        return model

    def validate_voice(self, voice: str) -> str:
        """
        Validate and return the voice name.

        Args:
            voice (str): Voice name to validate

        Returns:
            str: Validated voice name

        Raises:
            ValueError: If voice is not supported
        """
        if voice not in self.SUPPORTED_VOICES:
            raise ValueError(f"Voice '{voice}' not supported. Available voices: {', '.join(self.SUPPORTED_VOICES)}")
        return voice

    def validate_format(self, response_format: str) -> str:
        """
        Validate and return the response format.

        Args:
            response_format (str): Response format to validate

        Returns:
            str: Validated response format

        Raises:
            ValueError: If format is not supported
        """
        if response_format not in self.SUPPORTED_FORMATS:
            raise ValueError(f"Format '{response_format}' not supported. Available formats: {', '.join(self.SUPPORTED_FORMATS)}")
        return response_format

    def save_audio(self, audio_file: str, destination: str = None, verbose: bool = False) -> str:
        """
        Save audio to a specific destination.

        Args:
            audio_file (str): Path to the source audio file
            destination (str, optional): Destination path. Defaults to current directory with timestamp.
            verbose (bool, optional): Whether to print debug information. Defaults to False.

        Returns:
            str: Path to the saved audio file

        Raises:
            FileNotFoundError: If the audio file doesn't exist
        """
        import shutil
        import time

        source_path = Path(audio_file)

        if not source_path.exists():
            raise FileNotFoundError(f"Audio file not found: {audio_file}")

        if destination is None:
            # Create a default destination with timestamp in current directory
            timestamp = int(time.time())
            destination = os.path.join(os.getcwd(), f"speech_{timestamp}{source_path.suffix}")

        # Ensure the destination directory exists
        os.makedirs(os.path.dirname(os.path.abspath(destination)), exist_ok=True)

        # Copy the file
        shutil.copy2(source_path, destination)

        if verbose:
            ic.configureOutput(prefix='DEBUG| ')
            ic(f"Audio saved to {destination}")

        return destination

    def create_speech(
        self,
        input_text: str,
        model: str = None,
        voice: str = None,
        response_format: str = None,
        instructions: str = None,
        verbose: bool = False
    ) -> str:
        """
        Create speech from input text (OpenAI-compatible interface).

        Args:
            input_text (str): The text to convert to speech
            model (str, optional): The TTS model to use
            voice (str, optional): The voice to use
            response_format (str, optional): Audio format (mp3, opus, aac, flac, wav, pcm)
            instructions (str, optional): Voice instructions for controlling speech aspects
            verbose (bool, optional): Whether to print debug information

        Returns:
            str: Path to the generated audio file
        """
        # Use defaults if not provided
        model = model or self.default_model
        voice = voice or self.default_voice
        response_format = response_format or self.default_format

        # Validate parameters
        self.validate_model(model)
        self.validate_voice(voice)
        self.validate_format(response_format)

        # Call the provider-specific TTS implementation
        return self.tts(
            text=input_text,
            model=model,
            voice=voice,
            response_format=response_format,
            instructions=instructions,
            verbose=verbose
        )

    def stream_audio(
        self,
        input_text: str,
        model: str = None,
        voice: str = None,
        response_format: str = None,
        instructions: str = None,
        chunk_size: int = 1024,
        verbose: bool = False
    ) -> Generator[bytes, None, None]:
        """
        Stream audio in chunks with OpenAI-compatible parameters.

        Args:
            input_text (str): The text to convert to speech
            model (str, optional): The TTS model to use
            voice (str, optional): The voice to use
            response_format (str, optional): Audio format
            instructions (str, optional): Voice instructions
            chunk_size (int, optional): Size of audio chunks to yield. Defaults to 1024.
            verbose (bool, optional): Whether to print debug information. Defaults to False.

        Yields:
            Generator[bytes, None, None]: Audio data chunks
        """
        # Generate the audio file using create_speech
        audio_file = self.create_speech(
            input_text=input_text,
            model=model,
            voice=voice,
            response_format=response_format,
            instructions=instructions,
            verbose=verbose
        )

        # Stream the file in chunks
        with open(audio_file, 'rb') as f:
            while chunk := f.read(chunk_size):
                yield chunk

    def tts(self, text: str, **kwargs) -> str:
        """
        Abstract method for text-to-speech conversion.
        Must be implemented by subclasses.

        Args:
            text (str): The text to convert to speech
            **kwargs: Additional provider-specific parameters

        Returns:
            str: Path to the generated audio file

        Raises:
            NotImplementedError: If not implemented by subclass
        """
        raise NotImplementedError("Subclasses must implement the tts method")


class AsyncBaseTTSProvider:
    """
    Base class for async TTS providers with OpenAI-compatible functionality.

    This class implements common async methods following OpenAI TTS API patterns
    for speech generation, streaming, and audio handling.
    """
    required_auth = False

    # Supported models (can be overridden by subclasses)
    SUPPORTED_MODELS = [
        "gpt-4o-mini-tts",  # Latest intelligent realtime model
        "tts-1",            # Lower latency model
        "tts-1-hd"          # Higher quality model
    ]

    # Supported voices (can be overridden by subclasses)
    SUPPORTED_VOICES = [
        "alloy", "ash", "ballad", "coral", "echo",
        "fable", "nova", "onyx", "sage", "shimmer"
    ]

    # Supported output formats
    SUPPORTED_FORMATS = [
        "mp3",    # Default format
        "opus",   # Internet streaming, low latency
        "aac",    # Digital audio compression
        "flac",   # Lossless compression
        "wav",    # Uncompressed, low latency
        "pcm"     # Raw samples, 24kHz 16-bit
    ]

    def __init__(self):
        """Initialize the async base TTS provider."""
        self.temp_dir = tempfile.mkdtemp(prefix="webscout_tts_")
        self.default_model = "gpt-4o-mini-tts"
        self.default_voice = "coral"
        self.default_format = "mp3"

    async def validate_model(self, model: str) -> str:
        """
        Validate and return the model name.

        Args:
            model (str): Model name to validate

        Returns:
            str: Validated model name

        Raises:
            ValueError: If model is not supported
        """
        # If provider doesn't support models, return the model as-is
        if self.SUPPORTED_MODELS is None:
            return model

        if model not in self.SUPPORTED_MODELS:
            raise ValueError(f"Model '{model}' not supported. Available models: {', '.join(self.SUPPORTED_MODELS)}")
        return model

    async def validate_voice(self, voice: str) -> str:
        """
        Validate and return the voice name.

        Args:
            voice (str): Voice name to validate

        Returns:
            str: Validated voice name

        Raises:
            ValueError: If voice is not supported
        """
        if voice not in self.SUPPORTED_VOICES:
            raise ValueError(f"Voice '{voice}' not supported. Available voices: {', '.join(self.SUPPORTED_VOICES)}")
        return voice

    async def validate_format(self, response_format: str) -> str:
        """
        Validate and return the response format.

        Args:
            response_format (str): Response format to validate

        Returns:
            str: Validated response format

        Raises:
            ValueError: If format is not supported
        """
        if response_format not in self.SUPPORTED_FORMATS:
            raise ValueError(f"Format '{response_format}' not supported. Available formats: {', '.join(self.SUPPORTED_FORMATS)}")
        return response_format

    async def save_audio(self, audio_file: str, destination: str = None, verbose: bool = False) -> str:
        """
        Save audio to a specific destination asynchronously.

        Args:
            audio_file (str): Path to the source audio file
            destination (str, optional): Destination path. Defaults to current directory with timestamp.
            verbose (bool, optional): Whether to print debug information. Defaults to False.

        Returns:
            str: Path to the saved audio file

        Raises:
            FileNotFoundError: If the audio file doesn't exist
        """
        import asyncio
        import shutil
        import time

        source_path = Path(audio_file)

        if not source_path.exists():
            raise FileNotFoundError(f"Audio file not found: {audio_file}")

        if destination is None:
            # Create a default destination with timestamp in current directory
            timestamp = int(time.time())
            destination = os.path.join(os.getcwd(), f"speech_{timestamp}{source_path.suffix}")

        # Ensure the destination directory exists
        os.makedirs(os.path.dirname(os.path.abspath(destination)), exist_ok=True)

        # Copy the file using asyncio to avoid blocking
        await asyncio.to_thread(shutil.copy2, source_path, destination)

        if verbose:
            ic.configureOutput(prefix='DEBUG| ')
            ic(f"Audio saved to {destination}")

        return destination

    async def create_speech(
        self,
        input_text: str,
        model: str = None,
        voice: str = None,
        response_format: str = None,
        instructions: str = None,
        verbose: bool = False
    ) -> str:
        """
        Create speech from input text asynchronously (OpenAI-compatible interface).

        Args:
            input_text (str): The text to convert to speech
            model (str, optional): The TTS model to use
            voice (str, optional): The voice to use
            response_format (str, optional): Audio format (mp3, opus, aac, flac, wav, pcm)
            instructions (str, optional): Voice instructions for controlling speech aspects
            verbose (bool, optional): Whether to print debug information

        Returns:
            str: Path to the generated audio file
        """
        # Use defaults if not provided
        model = model or self.default_model
        voice = voice or self.default_voice
        response_format = response_format or self.default_format

        # Validate parameters
        await self.validate_model(model)
        await self.validate_voice(voice)
        await self.validate_format(response_format)

        # Call the provider-specific TTS implementation
        return await self.tts(
            text=input_text,
            model=model,
            voice=voice,
            response_format=response_format,
            instructions=instructions,
            verbose=verbose
        )

    async def stream_audio(
        self,
        input_text: str,
        model: str = None,
        voice: str = None,
        response_format: str = None,
        instructions: str = None,
        chunk_size: int = 1024,
        verbose: bool = False
    ):
        """
        Stream audio in chunks asynchronously with OpenAI-compatible parameters.

        Args:
            input_text (str): The text to convert to speech
            model (str, optional): The TTS model to use
            voice (str, optional): The voice to use
            response_format (str, optional): Audio format
            instructions (str, optional): Voice instructions
            chunk_size (int, optional): Size of audio chunks to yield. Defaults to 1024.
            verbose (bool, optional): Whether to print debug information. Defaults to False.

        Yields:
            AsyncGenerator[bytes, None]: Audio data chunks
        """
        try:
            import aiofiles
        except ImportError:
            raise ImportError("The 'aiofiles' package is required for async streaming. Install it with 'pip install aiofiles'.")

        # Generate the audio file using create_speech
        audio_file = await self.create_speech(
            input_text=input_text,
            model=model,
            voice=voice,
            response_format=response_format,
            instructions=instructions,
            verbose=verbose
        )

        # Stream the file in chunks
        async with aiofiles.open(audio_file, 'rb') as f:
            while chunk := await f.read(chunk_size):
                yield chunk

    async def tts(self, text: str, **kwargs) -> str:
        """
        Abstract async method for text-to-speech conversion.
        Must be implemented by subclasses.

        Args:
            text (str): The text to convert to speech
            **kwargs: Additional provider-specific parameters

        Returns:
            str: Path to the generated audio file

        Raises:
            NotImplementedError: If not implemented by subclass
        """
        raise NotImplementedError("Subclasses must implement the async tts method")



import warnings
from json import dumps
from os import path
from typing import Any, Dict, Generator, Optional, Union

from ..AIbase import Provider, Response

# Import internal modules and dependencies
from ..AIutel import Conversation, Optimizers
from ..Bard import Chatbot, Model

warnings.simplefilter("ignore", category=UserWarning)

# Define model aliases for easy usage (only supported models)
MODEL_ALIASES: Dict[str, Model] = {
    "unspecified": Model.UNSPECIFIED,
    "gemini-2.5-flash": Model.G_2_5_FLASH,
    "gemini-2.5-pro": Model.G_2_5_PRO,
    "gemini-3-pro": Model.G_3_PRO,
    "flash-2.5": Model.G_2_5_FLASH,
    "pro": Model.G_2_5_PRO,
}

# List of available models (friendly names)
AVAILABLE_MODELS = list(MODEL_ALIASES.keys())

class GEMINI(Provider):
    required_auth = True
    def __init__(
        self,
        cookie_file: str,
        model: str = "gemini-2.5-flash",  # Accepts either a Model enum or a str alias.
        proxy: dict = {},
        timeout: int = 30,
    ):
        """
        Initializes GEMINI with model support.

        Args:
            cookie_file (str): Path to the cookies JSON file.
            model (Model or str): Selected model for the session. Can be a Model enum
                or a string alias. Available aliases: flash, flash-exp, thinking, thinking-with-apps,
                exp-advanced, 2.5-exp-advanced, 2.5-pro, 1.5-flash, 1.5-pro, 1.5-pro-research.
            proxy (dict, optional): HTTP request proxy. Defaults to {}.
            timeout (int, optional): HTTP request timeout in seconds. Defaults to 30.
        """
        self.conversation = Conversation(False)

        # Ensure cookie_file existence.
        if not isinstance(cookie_file, str):
            raise TypeError(f"cookie_file should be of type str, not '{type(cookie_file)}'")
        if not path.isfile(cookie_file):
            raise Exception(f"{cookie_file} is not a valid file path")

        # If model is provided as alias (str), convert to Model enum.
        if isinstance(model, str):
            alias = model.lower()
            if alias in MODEL_ALIASES:
                selected_model = MODEL_ALIASES[alias]
            else:
                raise Exception(f"Unknown model alias: '{model}'. Available aliases: {', '.join(AVAILABLE_MODELS)}")
        elif isinstance(model, Model):
            selected_model = model
        else:
            raise TypeError("model must be a string alias or an instance of Model")

        # Initialize the Chatbot session using the cookie file.
        self.session = Chatbot(cookie_file, proxy, timeout, selected_model)
        self.last_response = {}
        self.__available_optimizers = (
            method for method in dir(Optimizers) if callable(getattr(Optimizers, method)) and not method.startswith("__")
        )
        # Store cookies from Chatbot for later use (e.g. image generation)
        self.session_auth1 = self.session.secure_1psid
        self.session_auth2 = self.session.secure_1psidts

    def ask(
        self,
        prompt: str,
        stream: bool = False,
        raw: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        **kwargs: Any,
    ) -> Response:
        """Chat with AI.

        Args:
            prompt (str): Prompt to be sent.
            stream (bool, optional): Flag for streaming response. Defaults to False.
            raw (bool, optional): Stream back raw response as received. Defaults to False.
            optimizer (str, optional): Prompt optimizer name (e.g., 'code', 'shell_command'). Defaults to None.
            conversationally (bool, optional): Chat conversationally when using optimizer. Defaults to False.

        Returns:
            dict: Response generated by the underlying Chatbot.
        """
        conversation_prompt = self.conversation.gen_complete_prompt(prompt)
        if optimizer:
            if optimizer in self.__available_optimizers:
                conversation_prompt = getattr(Optimizers, optimizer)(
                    conversation_prompt if conversationally else prompt
                )
            else:
                raise Exception(f"Optimizer is not one of {', '.join(self.__available_optimizers)}")

        def for_stream():
            response = self.session.ask(prompt)
            self.last_response.update(response)
            self.conversation.update_chat_history(prompt, self.get_message(self.last_response))
            if raw:
                yield dumps(response)
            else:
                yield response

        def for_non_stream():
            for _ in for_stream():
                pass
            return self.last_response if not raw else dumps(self.last_response)

        return for_stream() if stream else for_non_stream()

    def chat(
        self,
        prompt: str,
        stream: bool = False,
        optimizer: Optional[str] = None,
        conversationally: bool = False,
        raw: bool = False,
    ) -> Union[str, Generator[str, None, None]]:
        """Generate response text.

        Args:
            prompt (str): Prompt to be sent.
            stream (bool, optional): Flag for streaming response. Defaults to False.
            optimizer (str, optional): Prompt optimizer name. Defaults to None.
            conversationally (bool, optional): Chat conversationally when using optimizer. Defaults to False.
            raw (bool, optional): Return raw response. Defaults to False.

        Returns:
            str: Response generated.
        """
        def for_stream():
            for response in self.ask(prompt, True, raw=raw, optimizer=optimizer, conversationally=conversationally):
                if raw:
                    yield response
                else:
                    yield self.get_message(response)

        def for_non_stream():
            result = self.ask(prompt, False, raw=raw, optimizer=optimizer, conversationally=conversationally)
            if raw:
                return result
            return self.get_message(result)

        return for_stream() if stream else for_non_stream()

    def get_message(self, response: Response) -> str:
        """Retrieves message content from the response.

        Args:
            response (Response): Response generated by `self.ask`.

        Returns:
            str: Extracted message content.
        """
        if not isinstance(response, dict):
            return str(response)
        return response["content"]

    def reset(self):
        """Reset the current conversation."""
        self.session.async_chatbot.conversation_id = ""
        self.session.async_chatbot.response_id = ""
        self.session.async_chatbot.choice_id = ""
